from typing import Mapping
from typing import Union
from typing import List
from typing import Any
from typing import Callable
from typing import Optional
from typing import Dict
from collections.abc import Iterable

import abc
import copy
import joblib
import json
import yaml
import fsspec
import pandas as pd
import datamol as dm
import numpy as np

from rdkit.Chem import rdchem
from sklearn.base import TransformerMixin
from sklearn.base import BaseEstimator
from loguru import logger

from molfeat._version import __version__ as MOLFEAT_VERSION

from molfeat.calc import get_calculator
from molfeat.calc.base import _CALCULATORS
from molfeat.utils import datatype
from molfeat.utils.cache import _Cache, FileCache, MPDataCache
from molfeat.utils.cache import CacheList
from molfeat.utils.commons import fn_to_hex
from molfeat.utils.commons import hex_to_fn
from molfeat.utils.parsing import get_input_args
from molfeat.utils.parsing import import_from_string
from molfeat.utils.state import DTYPES_MAPPING
from molfeat.utils.state import DTYPES_MAPPING_REVERSE
from molfeat.utils.state import ATOM_FEATURIZER_MAPPING
from molfeat.utils.state import BOND_FEATURIZER_MAPPING
from molfeat.utils.state import ATOM_FEATURIZER_MAPPING_REVERSE
from molfeat.utils.state import BOND_FEATURIZER_MAPPING_REVERSE
from molfeat.utils.state import get_type_mapping

_TRANSFORMERS = {}


class _TransformerMeta(abc.ABCMeta):
    """Metaclass to register all transformers automatically"""

    def __init__(cls, name, bases, attrs):
        type.__init__(cls, name, bases, attrs)
        if name in _TRANSFORMERS.keys():
            logger.warning(
                f"The {name!r} interaction has been superseded by a "
                f"new class with id {id(cls):#x}"
            )
        # ignore private classes
        if name not in ["BaseFeaturizer"] and not name.startswith("_"):
            # do not register the base class
            _TRANSFORMERS[name] = cls


class BaseFeaturizer(BaseEstimator):
    """
    Molecule featurizer base class that needs to be implemented by all featurizers.
    This featurizer is compatible with scikit-learn estimators and thus can be plugged into a pipeline
    """

    def __init__(
        self,
        n_jobs: int = 1,
        verbose: bool = True,
        dtype: Optional[Union[str, Callable]] = None,
        parallel_kwargs: Optional[Dict[str, Any]] = None,
        **params,
    ):
        self._n_jobs = n_jobs
        self.dtype = dtype
        self.verbose = verbose
        self.parallel_kwargs = parallel_kwargs or {}
        for k, v in params.items():
            setattr(self, k, v)
        self._input_params = dict(n_jobs=n_jobs, dtype=dtype, verbose=verbose, **params)

    @property
    def n_jobs(self):
        """Get the number of concurrent jobs to run with this featurizer"""
        return self._n_jobs

    @n_jobs.setter
    def n_jobs(self, val):
        if val >= 1:
            self._n_jobs = val
        elif val == -1:
            self._n_jobs = joblib.cpu_count()

    def _get_param_names(self):
        """Get parameter names for the estimator"""
        return self._input_params.keys()

    def _update_params(self):
        """Update parameters of the current estimator"""
        ...

    def set_params(self, **params):
        """Set the parameters of this estimator.

        Returns:
            self: estimator instance
        """
        super().set_params(**params)
        for k, v in params.items():
            if k in self._input_params:
                self._input_params[k] = v
        self._update_params()
        return self

    def copy(self):
        """Return a copy of this object."""
        copy_obj = self.__class__(**self._input_params)
        for k, v in self.__dict__.items():
            if not hasattr(copy_obj, k):
                setattr(copy_obj, k, copy.deepcopy(v))
        return copy_obj

    def preprocess(self, inputs: list, labels: Optional[list] = None):
        """Preprocess input

        Args:
            inputs: inputs to preprocess
            labels: labels to preprocess (optional)

        Returns:
            processed: pre-processed input list
        """
        return inputs, labels

    def get_collate_fn(self, *args, **kwargs):
        """
        Get collate function of this featurizer. In the implementation of this function
        you should set the relevant attributes or argument of the underlying collate function
        (e.g via functools.partial) and return the function itself

        Returns:
            fn: Collate function for pytorch or None
        """
        return None


class MoleculeTransformer(TransformerMixin, BaseFeaturizer, metaclass=_TransformerMeta):

    """
    Base class for molecular data transformer such as Fingerprinter etc.
    If you create a subclass of this featurizer, you will need to make sure that the
    input argument of the init are kept as is in the object attributes.

    !!! note
        The transformer supports a variety of datatype, they are only enforced when passing the
        `enforce_dtype=True` attributes in `__call__`. For pandas dataframes, use `'pandas'|'df'|'dataframe'|pd.DataFrame`
    """

    def __init__(
        self,
        featurizer: Union[str, Callable],
        n_jobs: int = 1,
        verbose: bool = False,
        dtype: Optional[Union[str, Callable]] = None,
        parallel_kwargs: Optional[Dict[str, Any]] = None,
        **params,
    ):
        """Mol transformer base class

        Args:
            featurizer: featurizer to use
            n_jobs (int, optional): Number of job to run in parallel. Defaults to 1.
            verbose (bool, optional): Verbosity level. Defaults to True.
            dtype (callable, optional): Output data type. Defaults to None, where numpy arrays are returned.
            parallel_kwargs (dict, optional): Optional kwargs to pass to the dm.parallelized function. Defaults to None.

        """
        super().__init__(
            n_jobs=n_jobs,
            verbose=verbose,
            dtype=dtype,
            featurizer=featurizer,
            parallel_kwargs=parallel_kwargs,
            **params,
        )
        if callable(featurizer):
            self.featurizer = featurizer
        else:
            self.featurizer = get_calculator(featurizer, **params)

        self.cols_to_keep = None
        self._fitted = False

        self._save_input_args()

    def _save_input_args(self):
        """Save the input arguments of a transformer to the attribute
        `_input_args` of the object.
        """

        # NOTE(hadim): don't override existing _input_args so
        # it's possible to use MoleculeTransformer as a featurizer
        # instead of simply a base class.
        if not hasattr(self, "_input_args"):
            self._input_args = get_input_args()

    def _update_params(self):
        if not callable(self.featurizer):
            params = copy.deepcopy(self._input_params)
            params.pop("featurizer")
            self.featurizer = get_calculator(self.featurizer, **params)
        self._fitted = False

    def __setstate__(self, state):
        state.pop("callbacks", None)
        self.__dict__.update(state)
        self.__dict__["parallel_kwargs"] = state.get("parallel_kwargs", {})
        self._update_params()

    def fit(self, X: List[Union[rdchem.Mol, str]], y: Optional[list] = None, **fit_params):
        """Fit the current transformer on given dataset.

        The goal of fitting is for example to identify nan columns values
        that needs to be removed from the dataset

        Args:
            X: input list of molecules
            y (list, optional): Optional list of molecular properties. Defaults to None.

        Returns:
            self: MolTransformer instance after fitting
        """
        feats = self.transform(X, ignore_errors=True)
        lengths = [len(x) for x in feats if not datatype.is_null(x)]
        if lengths:
            # we will ignore all nan
            feats = datatype.to_numpy([f for f in feats if not datatype.is_null(f)])
            self.cols_to_keep = (~np.any(np.isnan(feats), axis=0)).nonzero()[0]
        self._fitted = True
        return self

    def _transform(self, mol: rdchem.Mol):
        r"""
        Compute features for a single molecule.
        This method would potentially need to be reimplemented by child classes

        Args:
            mol (rdchem.Mol): molecule to transform into features

        Returns
            feat: featurized input molecule

        """
        feat = None
        try:
            feat = datatype.to_numpy(self.featurizer(mol))
            if self.cols_to_keep is not None:
                feat = feat[self.cols_to_keep]
        except Exception as e:
            if self.verbose:
                logger.error(e)
        return feat

    def transform(
        self,
        mols: List[Union[rdchem.Mol, str]],
        ignore_errors: bool = False,
        **kwargs,
    ):
        r"""
        Compute the features for a set of molecules.

        !!! note
            Note that depending on the `ignore_errors` argument, all failed
            featurization (caused whether by invalid smiles or error during
            data transformation) will be substitued by None features for the
            corresponding molecule. This is done, so you can find the positions
            of these molecules and filter them out according to your own logic.

        Args:
            mols: a list containing smiles or mol objects
            ignore_errors (bool, optional): Whether to silently ignore errors


        Returns:
            features: a list of features for each molecule in the input set
        """
        # Convert single mol to iterable format
        if isinstance(mols, (str, rdchem.Mol)) or not isinstance(mols, Iterable):
            mols = [mols]

        def _to_mol(x):
            return dm.to_mol(x) if x else None

        parallel_kwargs = getattr(self, "parallel_kwargs", {})

        mols = dm.parallelized(_to_mol, mols, n_jobs=self.n_jobs, **parallel_kwargs)

        if self.n_jobs not in [0, 1]:
            # use a proxy model to run in parallel
            cpy = self.copy()
            features = dm.parallelized(
                cpy._transform,
                mols,
                n_jobs=self.n_jobs,
                **cpy.parallel_kwargs,
            )
        else:
            features = [self._transform(mol) for mol in mols]
        if not ignore_errors:
            for ind, feat in enumerate(features):
                if feat is None:
                    raise ValueError(f"Cannot transform molecule at index {ind}")

        return features

    def __len__(self):
        """Compute featurizer length"""

        # check length and _length attribute
        cols_to_keep = getattr(self, "cols_to_keep", None)
        cur_length = None

        if cols_to_keep is not None:
            cur_length = len(cols_to_keep)
        else:
            cur_length = getattr(self, "length", getattr(self, "_length", None))
            # then check the featurizer length if it's a callable and not a string/None
            if (
                cur_length is None
                and callable(self.featurizer)
                and hasattr(self.featurizer, "__len__")
            ):
                cur_length = len(self.featurizer)

        if cur_length is None:
            raise ValueError(
                f"Cannot auto-determine length of this MolTransformer: {self.__class__.__name__}"
            )

        return cur_length

    def __call__(
        self,
        mols: List[Union[rdchem.Mol, str]],
        enforce_dtype: bool = True,
        ignore_errors: bool = False,
        **kwargs,
    ):
        r"""
        Calculate features for molecules. Using __call__, instead of transform.
        If ignore_error is True, a list of features and valid ids are returned.
        Note that most Transfomers allow you to specify
        a return datatype.

        Args:
            mols:  Mol or SMILES of the molecules to be transformed
            enforce_dtype: whether to enforce the instance dtype in the generated fingerprint
            ignore_errors: Whether to ignore errors during featurization or raise an error.
            kwargs: Named parameters for the transform method

        Returns:
            feats: list of valid features
            ids: all valid molecule positions that did not failed during featurization.
                Only returned when ignore_errors is True.

        """
        features = self.transform(mols, ignore_errors=ignore_errors, enforce_dtype=False, **kwargs)
        ids = np.arange(len(features))
        if ignore_errors:
            features, ids = self._filter_none(features)
        if self.dtype is not None and enforce_dtype:
            features = datatype.cast(features, dtype=self.dtype, columns=self.columns)
        if not ignore_errors:
            return features
        return features, ids

    @staticmethod
    def _filter_none(features):
        ids_bad = []
        # If the features are a list, filter the None ids
        if isinstance(features, (tuple, list, np.ndarray)):
            for f_id, feat in enumerate(features):
                if datatype.is_null(feat):
                    ids_bad.append(f_id)
            ids_to_keep = [
                this_id for this_id in np.arange(0, len(features)) if this_id not in ids_bad
            ]
            features = [features[ii] for ii in ids_to_keep]

        # If the features are a dict or DataFrame, filter the ids when any key id is None
        elif isinstance(features, (dict, pd.DataFrame)):
            if isinstance(features, dict):
                features = pd.DataFrame(features)
            for feat_col in features.columns:
                for f_id, feat in enumerate(features[feat_col].values.flatten()):
                    if feat is None:
                        ids_bad.append(f_id)
            ids_bad = np.unique(ids_bad).tolist()
            all_ids = np.arange(0, features.shape[0])
            ids_to_keep = [this_id for this_id in all_ids if this_id not in ids_bad]
            features = features.iloc[ids_to_keep, :]

        else:
            ids_to_keep = np.arange(0, features.shape[0])
        return features, list(ids_to_keep)

    @property
    def columns(self):
        """Get the list of columns for this molecular descriptor

        Returns:
            columns (list): Name of the columns of the descriptor
        """
        columns = getattr(self.featurizer, "columns", None)
        cols_to_keep = getattr(self, "cols_to_keep", None)
        if columns is not None and cols_to_keep is not None and len(cols_to_keep) > 0:
            columns = [columns[i] for i in cols_to_keep]
        return columns

    @staticmethod
    def batch_transform(
        transformer: Callable,
        mols: List[Union[rdchem.Mol, str]],
        batch_size: int = 256,
        n_jobs: Optional[int] = None,
        concatenate: bool = True,
        progress: bool = True,
        leave_progress: bool = False,
        **parallel_kwargs,
    ):
        """
        Batched computation of featurization of a list of molecules

        Args:
            transformer: Fingerprint transformer
            mols: List of molecules to featurize
            batch_size: Batch size
            n_jobs: number of jobs to run in parallel
            concatenate: Whether to concatenate the results or return the list of batched results
            progress: whether to show progress bar
            leave_progress: whether to leave progress bar after completion
            parallel_kwargs: additional arguments to pass to dm.parallelized

        Returns:
            List of batches
        """

        step_size = int(np.ceil(len(mols) / batch_size))
        batched_mols = np.array_split(mols, step_size)

        tqdm_kwargs = parallel_kwargs.setdefault("tqdm_kwargs", {})
        tqdm_kwargs.update(leave=leave_progress, desc="Batch compute:")
        parallel_kwargs["tqdm_kwargs"] = tqdm_kwargs

        # it's recommended to use a precomputed molecule transformer
        # instead of the internal cache for pretrained models
        cache_attr = "cache"
        existing_cache = getattr(transformer, cache_attr, None)
        if existing_cache is None:
            cache_attr = "precompute_cache"
            existing_cache = getattr(transformer, cache_attr, None)

        use_mp_cache = (
            existing_cache is not None
            and not isinstance(existing_cache, MPDataCache)
            and n_jobs not in [None, 0, 1]  # this is based on datamol sequential vs parallel
        )
        if use_mp_cache:
            # we need to change the cache system to one that works with multiprocessing
            # to have a shared memory
            new_cache = MPDataCache()
            new_cache.update(existing_cache)
            transformer.cache = new_cache

        transformed = dm.parallelized(
            transformer,
            batched_mols,
            n_jobs=n_jobs,
            progress=progress,
            **parallel_kwargs,
        )
        if use_mp_cache:
            # we set back the original transformation while updating it with
            # all the missing values
            existing_cache.update(transformer.cache)
            setattr(transformer, cache_attr, existing_cache)

        if concatenate:
            # if we ask for concatenation, then we would need to fix None values ideally
            fixed_transformations = []
            for computed_trans in transformed:
                if computed_trans is None:
                    computed_trans = np.full(len(computed_trans), len(transformer), np.nan)
                else:
                    for i, x in enumerate(computed_trans):
                        if x is None:
                            computed_trans[i] = np.full(len(transformer), np.nan)
                fixed_transformations.append(computed_trans)
            return np.concatenate(fixed_transformations)
        return transformed

    # Featurizer to state methods

    def to_state_dict(self) -> dict:
        """Serialize the featurizer to a state dict."""

        if getattr(self, "_input_args") is None:
            raise ValueError(f"Cannot save state for this transformer '{self.__class__.__name__}'")

        # Process the input arguments before building the state
        args = copy.deepcopy(self._input_args)

        ## Deal with dtype
        if "dtype" in args and not isinstance(args["dtype"], str):
            if args["dtype"] not in DTYPES_MAPPING:
                raise ValueError(
                    f"Invalid dtype {args['dtype']}. Valid dtypes are {DTYPES_MAPPING.keys()}"
                )
            args["dtype"] = DTYPES_MAPPING[args["dtype"]]

        ## Deal with graph atom/bond featurizers
        # NOTE(hadim): it's important to highlight that atom/bond featurizers can't be
        # customized with this logic.
        if args.get("atom_featurizer") is not None:
            if hasattr(args.get("atom_featurizer"), "to_state_dict"):
                args["atom_featurizer"] = args["atom_featurizer"].to_state_dict()
                args["_atom_featurizer_is_pickled"] = False
            else:
                logger.warning
                (
                    "You are attempting to pickle an atom featurizer without a `to_state_dict` function into a hex string"
                )
                args["atom_featurizer"] = fn_to_hex(args["atom_featurizer"])
                args["_atom_featurizer_is_pickled"] = True

        # deal with bond featurizer
        if args.get("bond_featurizer") is not None:
            if hasattr(args.get("bond_featurizer"), "to_state_dict"):
                args["bond_featurizer"] = args["bond_featurizer"].to_state_dict()
                args["_bond_featurizer_is_pickled"] = False
            else:
                logger.warning(
                    "You are attempting to pickle a bond featurizer without a `to_state_dict` function into a hex string"
                )
                args["bond_featurizer"] = fn_to_hex(args["bond_featurizer"])
                args["_bond_featurizer_is_pickled"] = True

        ## Deal with custom featurizer
        if "featurizer" in args and isinstance(args["featurizer"], Callable):
            if hasattr(args["featurizer"], "to_state_dict"):
                args["featurizer"] = args["featurizer"].to_state_dict()
                args["_featurizer_is_pickled"] = False
            else:
                logger.warning(
                    "You are attempting to pickle a callable without a `to_state_dict` function into a hex string"
                )
                args["featurizer"] = fn_to_hex(args["featurizer"])
                args["_featurizer_is_pickled"] = True

        # Build the state
        state = {}
        state["name"] = self.__class__.__name__
        state["args"] = args
        state["_molfeat_version"] = MOLFEAT_VERSION
        return state

    def to_state_json(self) -> str:
        return json.dumps(self.to_state_dict())

    def to_state_yaml(self) -> str:
        return yaml.dump(self.to_state_dict(), Dumper=yaml.SafeDumper)

    def to_state_json_file(self, filepath: str):
        with fsspec.open(filepath, "w") as f:
            f.write(self.to_state_json())  # type: ignore

    def to_state_yaml_file(self, filepath: str):
        with fsspec.open(filepath, "w") as f:
            f.write(self.to_state_yaml())  # type: ignore

    # State to featurizer methods

    @staticmethod
    def from_state_dict(state: dict, override_args: Optional[dict] = None) -> "MoleculeTransformer":
        """Reload a featurizer from a state dict."""

        # Don't alter the original state dict
        state = copy.deepcopy(state)

        # MoleculeTransformer is a special case that has his own logic
        if state["name"] == "PrecomputedMolTransformer":
            return PrecomputedMolTransformer.from_state_dict(
                state=state,
                override_args=override_args,
            )

        # Get the name
        transformer_class = _TRANSFORMERS.get(state["name"])
        if transformer_class is None:
            raise ValueError(f"The featurizer '{state['name']}' is not supported.")
        if isinstance(transformer_class, str):
            # Get the transformer class from its path
            transformer_class = import_from_string(transformer_class)

        # Process the state as needed
        args = state.get("args", {})

        ## Deal with dtype
        if "dtype" in args:
            if args["dtype"] not in DTYPES_MAPPING_REVERSE:
                raise ValueError(
                    f"Invalid dtype {args['dtype']}. Valid dtypes are"
                    f" {DTYPES_MAPPING_REVERSE.keys()}"
                )
            args["dtype"] = DTYPES_MAPPING_REVERSE[args["dtype"]]

        ## Deal with graph atom/bond featurizers
        if args.get("atom_featurizer") is not None:
            if not args.get("_atom_featurizer_is_pickled"):
                klass_name = args["atom_featurizer"].get("name")
                args["atom_featurizer"] = ATOM_FEATURIZER_MAPPING_REVERSE[
                    klass_name
                ].from_state_dict(args["atom_featurizer"])
            else:
                # buffer = io.BytesIO(bytes.fromhex(args["atom_featurizer"]))
                # args["atom_featurizer"] = joblib.load(buffer)
                args["atom_featurizer"] = hex_to_fn(args["atom_featurizer"])
            args.pop("_atom_featurizer_is_pickled", None)
        if args.get("bond_featurizer") is not None:
            if not args.get("_bond_featurizer_is_pickled"):
                klass_name = args["bond_featurizer"].get("name")
                args["bond_featurizer"] = BOND_FEATURIZER_MAPPING_REVERSE[
                    klass_name
                ].from_state_dict(args["bond_featurizer"])
            else:
                # buffer = io.BytesIO(bytes.fromhex(args["bond_featurizer"]))
                # args["bond_featurizer"] = joblib.load(buffer)
                args["bond_featurizer"] = hex_to_fn(args["bond_featurizer"])
            args.pop("_bond_featurizer_is_pickled", None)
        ## Deal with custom featurizer
        if "featurizer" in args:
            if args.get("_featurizer_is_pickled") is True:
                # buffer = io.BytesIO(bytes.fromhex(args["featurizer"]))
                # args["featurizer"] = joblib.load(buffer)
                args["featurizer"] = hex_to_fn(args["featurizer"])
                args.pop("_featurizer_is_pickled")
            elif not isinstance(args["featurizer"], str):
                klass_name = args["featurizer"].get("name")
                args["featurizer"] = _CALCULATORS[klass_name].from_state_dict(args["featurizer"])
                args.pop("_featurizer_is_pickled")

        if override_args is not None:
            args.update(override_args)

        # Create the transformer
        featurizer = transformer_class(**args)
        return featurizer

    @staticmethod
    def from_state_json(
        state_json: str,
        override_args: Optional[dict] = None,
    ) -> "MoleculeTransformer":
        state_dict = json.loads(state_json)
        return MoleculeTransformer.from_state_dict(state_dict, override_args=override_args)

    @staticmethod
    def from_state_yaml(
        state_yaml: str,
        override_args: Optional[dict] = None,
    ) -> "MoleculeTransformer":
        state_dict = yaml.load(state_yaml, Loader=yaml.SafeLoader)
        return MoleculeTransformer.from_state_dict(state_dict, override_args=override_args)

    @staticmethod
    def from_state_json_file(
        filepath: str,
        override_args: Optional[dict] = None,
    ) -> "MoleculeTransformer":
        with fsspec.open(filepath, "r") as f:
            featurizer = MoleculeTransformer.from_state_json(f.read(), override_args=override_args)  # type: ignore
        return featurizer

    @staticmethod
    def from_state_yaml_file(
        filepath: str,
        override_args: Optional[dict] = None,
    ) -> "MoleculeTransformer":
        with fsspec.open(filepath, "r") as f:
            featurizer = MoleculeTransformer.from_state_yaml(f.read(), override_args=override_args)  # type: ignore
        return featurizer


class PrecomputedMolTransformer(MoleculeTransformer):
    """Convenience class for storing precomputed features."""

    def __init__(
        self,
        cache: Optional[Union[_Cache, Mapping[Any, Any], CacheList]] = None,
        cache_dict: Optional[Dict[str, Union[_Cache, Mapping[Any, Any], CacheList]]] = None,
        cache_key: Optional[str] = None,
        *args,
        featurizer: Optional[Union[MoleculeTransformer, str]] = None,
        state_path: Optional[str] = None,
        **kwargs,
    ):
        """
        Transformer that supports precomputation of features. You can either pass an empty cache or a pre-populated cache

        Args:
            cache: a datastructure of type mapping that maps each molecule to the precomputed features
            cache_dict: A dictionary of cache objects. This is a convenient structure when use multiple
                datacache for model selection.
            cache_key: The key of cache object to use.
            featurizer: optional featurizer used to compute the features of values not in the cache.
                Either the featurizer object or a string.
            state_path: optional state file path used to initiate the transformer object at the initialization
        """
        if (state_path is not None) and (
            (cache is not None) or (cache_dict is not None and cache_key is not None)
        ):
            raise ValueError(
                "`PrecomputedMolTransformer` can only be initiated by either `state_path` or"
                " the rest of parameters for cache and featurizer. But both are given."
            )

        super().__init__(*args, featurizer="none", **kwargs)

        if state_path is not None:
            self.__dict__ = self.from_state_file(state_path=state_path).__dict__.copy()
        else:
            if cache_dict is not None and cache_key is not None:
                self.cache_key = cache_key
                self.cache = cache_dict[self.cache_key]
            elif cache is not None:
                self.cache = cache
            else:
                raise AttributeError("The cache is not specified.")

            if isinstance(featurizer, str):
                self.base_featurizer = MoleculeTransformer(featurizer, *args, **kwargs)
            else:
                self.base_featurizer = featurizer

        # Set the length of the featurizer
        if len(self.cache) > 0:
            self.length = len(list(self.cache.values())[0])
        elif self.base_featurizer is not None:
            self.length = len(self.base_featurizer)
        else:
            raise AttributeError(
                "The cache is empty and the base featurizer is not specified. It's impossible"
                " to determine the length of the featurizer."
            )

    def _transform(self, mol: rdchem.Mol):
        r"""
        Return precomputed feature for a single molecule

        Args:
            mol (rdchem.Mol): molecule to transform into features

        Returns
            feat: featurized input molecule

        """
        feat = self.cache.get(mol)
        # if feat is None and we have an existing featurizer, we can update the cache
        if feat is None and self.base_featurizer is not None:
            feat = self.base_featurizer._transform(mol)
            self.cache[mol] = feat

        try:
            feat = datatype.to_numpy(feat)
            if self.cols_to_keep is not None:
                feat = feat[self.cols_to_keep]
        except Exception as e:
            if self.verbose:
                logger.error(e)
        return feat

    def update(self, feat_dict: Mapping[Any, Any]):
        r"""
        Fill the cache with new set of features for the molecules in mols.

        Args:
            feat_dict: A dictionary of molecules to features.
        """
        self.cache.update(feat_dict)

    def __getstate__(self):
        """Get the state for pickling"""
        state = {k: copy.deepcopy(v) for k, v in self.__dict__.items() if k not in ["cache"]}
        if isinstance(self.cache, FileCache):
            state["file_cache_args"] = dict(
                cache_file=self.cache.cache_file,
                name=self.cache.name,
                mol_hasher=self.cache.mol_hasher,
                n_jobs=self.cache.n_jobs,
                verbose=self.cache.verbose,
                file_type=self.cache.file_type,
                parquet_kwargs=self.cache.parquet_kwargs,
            )
        else:
            # EN: we do not copy the cache
            state["cache"] = self.cache
        return state

    def __setstate__(self, state):
        if "file_cache_args" in state:
            cache = FileCache(**state.pop("file_cache_args"))
            state["cache"] = cache
        return super().__setstate__(state)

    def to_state_dict(self, save_to_file: bool = True) -> dict:
        """Serialize a PrecomputedMolTransformer object to a state dict.

        Notes:
            - The base_featurizer must be set or a ValueError will be raised.
            - The cache must be a FileCache object or a ValueError will be raised.

        Args:
            save_to_file: whether to save the cache to file.
        """

        if self.base_featurizer is None:
            raise ValueError(
                "You can't serialize a PrecomputedMolTransformer that does not contain a"
                " featurizer."
            )

        if not isinstance(self.cache, FileCache):
            raise ValueError("The cache must be a FileCache object.")

        state = {}
        state["name"] = "PrecomputedMolTransformer"
        state["base_featurizer"] = self.base_featurizer.to_state_dict()
        state["cache"] = self.cache.to_state_dict(save_to_file=save_to_file)
        state["_molfeat_version"] = MOLFEAT_VERSION

        return state

    @staticmethod
    def from_state_dict(
        state: dict,
        override_args: Optional[dict] = None,
    ) -> "PrecomputedMolTransformer":
        # Don't alter the original state dict
        state = copy.deepcopy(state)

        args = {}

        # Load the FileCache object
        args["cache"] = FileCache.from_state_dict(state["cache"])

        # Load the base featurizer
        args["featurizer"] = MoleculeTransformer.from_state_dict(state["base_featurizer"])

        if override_args is not None:
            args.update(override_args)

        # Doesn't allow state_path in the initiation args
        args.pop("state_path", None)
        return PrecomputedMolTransformer(**args)

    def from_state_file(
        self,
        state_path: str,
        override_args: Optional[dict] = None,
    ) -> "PrecomputedMolTransformer":
        if state_path.endswith("yaml") or state_path.endswith("yml"):
            return self.from_state_yaml_file(filepath=state_path, override_args=override_args)
        elif state_path.endswith("json"):
            return self.from_state_json_file(filepath=state_path, override_args=override_args)
        else:
            raise ValueError(
                "Only files with 'yaml' or 'json' format are allowed. "
                "The filename must be ending with `yaml`, 'yml' or 'json'."
            )



import unittest as ut
import pytest
import torch
import numpy as np
from molfeat.utils import requires
from molfeat.trans.struct.esm import ESMProteinFingerprint
from molfeat.trans.struct.prot1D import ProtBioFingerprint


class Test_ESMProteinFingerprint(ut.TestCase):
    protein_name = "protein1"
    protein_seq = "MKTVRQERLKSIVRILERSKEPVSGAQLAEELSVSRQVIVQDIAYLRSLGYNIVATPRGYVLAGG"
    expected_dim = 768

    def test_esm_rep(self):
        transformer = ESMProteinFingerprint(featurizer="esm1_t6_43M_UR50S", pooling="max")
        fps = transformer.transform(self.protein_seq)
        self.assertEqual(len(fps), 1)
        self.assertEqual(fps[0].shape[-1], self.expected_dim)

        layers = [5, 6]
        transformer2 = ESMProteinFingerprint(
            featurizer="esm1_t6_43M_UR50S", layers=layers, pooling="max"
        )
        fps = transformer2.transform(self.protein_seq)
        self.assertEqual(len(fps), 1)
        self.assertEqual(fps[0].shape[-1], self.expected_dim * len(layers))

    def test_esm_token_rep(self):
        transformer = ESMProteinFingerprint(featurizer="esm1_t6_43M_UR50S", pooling=None)
        fps = transformer.transform(self.protein_seq)
        self.assertEqual(fps[0].shape[-1], self.expected_dim)
        self.assertEqual(fps[0].shape[0], len(self.protein_seq) + 1)

    def test_esm_contact(self):
        transformer = ESMProteinFingerprint(
            featurizer="esm1_t6_43M_UR50S", pooling="max", contact=True
        )
        fps = transformer.transform(self.protein_seq)
        self.assertTrue(fps[0].shape[0] == fps[0].shape[-1] == len(self.protein_seq))


@pytest.mark.xfail(
    not requires.check("bio_embeddings"), reason="3rd party module bio_embeddings is missing"
)
class TestProtBioFingerprint(ut.TestCase):
    protein_seq = [
        "MKTVRQERLKSIVRILERSKEPVSGAQLAEELSVSRQVIVQDIAYLRSLGYNIVATPRGYVLAGG",
        "KALTARQQEVFDLIRD",
    ]

    def test_unsupported_fp(self):
        with self.assertRaises(ValueError):
            transf = ProtBioFingerprint(featurizer="unsupported")
            transf(self.protein_seq)

    def test_unpooled_fp(self):
        trans1 = ProtBioFingerprint(featurizer="fasttext", pooling=None, device=torch.device("cpu"))
        trans2 = ProtBioFingerprint(featurizer="one_hot_encoding", pooling=None, device=None)
        out1 = trans1(self.protein_seq, enforce_dtype=True)
        out2 = trans2(self.protein_seq)
        self.assertIsInstance(out1, list)  # we expect list because we cannot concat
        self.assertEqual(len(out1), len(self.protein_seq))
        self.assertEqual(len(out1), len(out2))
        self.assertEqual(out1[0].shape[-1], len(trans1))

    def test_pooled_fp(self):
        transf = ProtBioFingerprint(
            featurizer="glove",
            pooling="mean",
            device=torch.device("cpu"),
            dtype=torch.float,
        )
        out1 = transf._transform(self.protein_seq[0])
        out2 = transf.transform(self.protein_seq)
        out3, _ = transf(self.protein_seq, enforce_dtype=True, ignore_errors=True)
        self.assertEqual(out3.shape[0], len(self.protein_seq))
        self.assertEqual(out3.shape[1], len(transf))
        self.assertEqual(len(out1.shape), 1)
        self.assertEqual(len(out2), len(self.protein_seq))
        self.assertTrue(torch.is_tensor(out3[0]))
        np.testing.assert_almost_equal(out2[0], out3[0].numpy().squeeze())



from typing import Optional
from typing import List
from typing import Union

from datetime import datetime
from pydantic.typing import Literal
from pydantic import BaseModel
from pydantic import Field
import datamol as dm


def get_model_init(card):
    """Get the model initialization code

    Args:
        card: model card to use
    """
    if card.group == "all" and card.type != "pretrained":
        import_statement = "from molfeat.trans import MoleculeTransformer"
        loader_statement = f"MoleculeTransformer(featurizer='{card.name}')"
    elif card.group in ["rdkit", "fp"]:
        import_statement = f"from molfeat.trans import FPVecTransformer"
        loader_statement = f"FPVecTransformer(kind='{card.name}')"
    elif card.group == "dgllife":
        import_statement = "from molfeat.trans.pretrained import PretrainedDGLTransformer"
        loader_statement = f"PretrainedDGLTransformer(kind='{card.name}')"
    elif card.group == "graphormer":
        import_statement = "from molfeat.trans.pretrained import GraphormerTransformer"
        loader_statement = f"GraphormerTransformer(kind='{card.name}')"
    elif card.group == "fcd":
        import_statement = "from molfeat.trans.pretrained import FCDTransformer"
        loader_statement = f"FCDTransformer()"
    elif card.group == "huggingface":
        import_statement = (
            "from molfeat.trans.pretrained.hf_transformers import PretrainedHFTransformer"
        )
        loader_statement = f"PretrainedHFTransformer(kind='{card.name}', notation='{card.inputs}')"
    else:
        raise ValueError(f"Unknown model group {card.group}")
    return import_statement, loader_statement


class ModelInfo(BaseModel):
    name: str
    inputs: str = "smiles"
    type: Literal["pretrained", "hand-crafted", "hashed", "count"]
    version: int = 0
    group: Optional[str] = "all"
    submitter: str
    description: str
    representation: Literal["graph", "line-notation", "vector", "tensor", "other"]
    require_3D: Optional[bool] = False
    tags: Optional[List[str]]
    authors: Optional[List[str]]
    reference: Optional[str]
    created_at: datetime = Field(default_factory=datetime.now)
    sha256sum: Optional[str]

    def path(self, root_path: str):
        """Generate the folder path where to save this model

        Args:
            root_path: path to the root folder
        """
        version = str(self.version or 0)
        return dm.fs.join(root_path, self.group, self.name, version)

    def match(self, new_card: Union["ModelInfo", dict], match_only: Optional[List[str]] = None):
        """Compare two model card information and returns True if they are the same

        Args:
            new_card: card to search for in the modelstore
            match_only: list of minimum attribute that should match between the two model information
        """

        self_content = self.dict().copy()
        if not isinstance(new_card, dict):
            new_card = new_card.dict()
        new_content = new_card.copy()
        # we always remove the datetime field
        self_content.pop("created_at", None)
        new_content.pop("created_at", None)
        if match_only is not None:
            self_content = {k: self_content.get(k) for k in match_only}
            new_content = {k: new_content.get(k) for k in match_only}
        return self_content == new_content

    def usage(self):
        """Return the usage of the model"""
        import_statement, loader_statement = get_model_init(self)
        usage = f"""
        import datamol as dm
        {import_statement}
        data = dm.freesolv().iloc[:100]
        transformer = {loader_statement}
        features = transformer(data["smiles"])
        """
        usage = "\n".join([x.strip() for x in usage.split("\n")])
        return usage



import numpy as np

from molfeat.trans.pretrained.base import PretrainedMolTransformer
from molfeat.utils import requires

if requires.check("fcd_torch"):
    from fcd_torch import FCD


class FCDTransformer(PretrainedMolTransformer):
    r"""
    FCD transformer based on the ChemNet pretrained model

    Attributes:
        featurizer (FCD): FCD featurizer object
        dtype (type, optional): Data type. Use call instead
    """

    def __init__(self, n_jobs=1, dtype=np.float32, **params):
        super().__init__(dtype=dtype, **params)
        if not requires.check("fcd_torch"):
            raise ImportError(
                "`fcd_torch` is not available, please install it `conda install -c conda-forge fcd_torch'`"
            )

        self.n_jobs = n_jobs
        self.featurizer = FCD(n_jobs=n_jobs)

    def _embed(self, smiles, **kwargs):
        """Compute embedding"""
        return self.featurizer.get_predictions(smiles)

    def _update_params(self):
        super()._update_params()
        self.featurizer = FCD(n_jobs=self.n_jobs)



from typing import List
from typing import Union
from typing import Callable
from typing import Optional

import os
import re
import tempfile
import fsspec
import joblib
import torch
import numpy as np
import datamol as dm
from loguru import logger
from torch.utils.data import DataLoader

from molfeat.utils.commons import _parse_to_evaluable_str
from molfeat.store.loader import PretrainedStoreModel
from molfeat.trans.pretrained.base import PretrainedMolTransformer
from molfeat.utils import requires
from molfeat.store import ModelStore

if requires.check("dgl") and requires.check("dgllife"):
    import dgl
    from dgl.nn.pytorch.glob import AvgPooling, MaxPooling, SumPooling

    from dgllife.utils import (
        mol_to_bigraph,
        PretrainAtomFeaturizer,
        PretrainBondFeaturizer,
        JTVAEVocab,
    )
    from dgllife.data import JTVAEDataset, JTVAECollator


class DGLModel(PretrainedStoreModel):
    r"""
    Load one of the pretrained DGL models for molecular embedding:
    """
    AVAILABLE_MODELS = [
        "gin_supervised_contextpred",
        "gin_supervised_infomax",
        "gin_supervised_edgepred",
        "gin_supervised_masking",
        "JTVAE_ZINC_no_kl",
    ]

    def __init__(
        self,
        name: str,
        cache_path: Optional[os.PathLike] = None,
        store: Optional[ModelStore] = None,
    ):
        if name not in self.AVAILABLE_MODELS:
            raise ValueError(f"{name} is not a supported pretrained gin model")

        super().__init__(name, cache_path=cache_path, store=store)
        self._model = None

    @classmethod
    def available_models(cls, query: Optional[str] = None):
        """List available models
        Args:
            query (str, optional): Query to filter the list of available models. Defaults to None.
        """
        if query is None:
            return cls.AVAILABLE_MODELS
        else:
            return [x for x in cls.AVAILABLE_MODELS if re.search(query, x, re.IGNORECASE)]

    @classmethod
    def from_pretrained(cls, model_name: str):
        """Load pretrained model using the dgllife API and not the store"""
        if not requires.check("dgllife"):
            raise ValueError("dgllife is not installed")
        import dgllife

        base_model = dgllife.model.load_pretrained(model_name)
        model = DGLModel(name=model_name)
        model.eval()
        model._model = base_model
        return model

    def load(self):
        """Load GIN model"""
        if self._model is not None:
            return self._model
        download_output_dir = self._artifact_load(
            name=self.name, download_path=self.cache_path, store=self.store
        )
        model_path = dm.fs.join(download_output_dir, self.store.MODEL_PATH_NAME)
        with fsspec.open(model_path, "rb") as f:
            model = joblib.load(f)
        model.eval()
        return model


class PretrainedDGLTransformer(PretrainedMolTransformer):
    r"""
    DGL Pretrained transformer

    Attributes:
        featurizer (DGLModel): DGL featurizer object
        dtype (type, optional): Data type.
        pooling (str, optional): Pooling method for GIN's embedding layer (Default: mean)
        batch_size (int, optional): Batch size to consider for model
    """

    def __init__(
        self,
        kind: Union[str, DGLModel] = "gin_supervised_contextpred",
        dtype: Callable = np.float32,
        pooling: str = "mean",
        batch_size: int = 32,
        preload: bool = False,
        **params,
    ):
        """DGL pretrained featurizer

        Args:
            kind (str, optional): name of the pretrained gin. Defaults to "gin_supervised_contextpred".
            dtype: datatype. Defaults to np.float32.
            pooling: global pooling to perform. Defaults to "mean".
            batch_size: batch size for featurizing the molecules. Defaults to 32.
            preload: whether to preload the internal pretrained featurizer or not

        """
        if not requires.check("dgllife"):
            raise ValueError("Cannot find dgl|dgllife. It's required for this featurizer !")
        super().__init__(
            dtype=dtype,
            pooling=pooling,
            batch_size=batch_size,
            preload=preload,
            kind=kind,
            **params,
        )
        self.pooling = pooling
        self.preload = preload
        self._pooling_obj = self.get_pooling(pooling)
        if isinstance(kind, DGLModel):
            self.kind = kind.name
            self.featurizer = kind
        else:
            self.kind = kind
            self.featurizer = DGLModel(name=self.kind)
        self.batch_size = int(batch_size)
        if self.preload:
            self._preload()

    def __repr__(self):
        return "{}(kind={}, pooling={}, dtype={})".format(
            self.__class__.__name__,
            _parse_to_evaluable_str(self.kind),
            _parse_to_evaluable_str(self.pooling),
            _parse_to_evaluable_str(self.dtype),
        )

    def _update_params(self):
        super()._update_params()
        self._pooling_obj = self.get_pooling(self.pooling)
        featurizer = DGLModel(name=self.kind)
        self.featurizer = featurizer.load()

    @staticmethod
    def get_pooling(pooling: str):
        """Get pooling method from name

        Args:
            pooling: name of the pooling method
        """
        pooling = pooling.lower()
        if pooling in ["mean", "avg", "average"]:
            return AvgPooling()
        elif pooling == "sum":
            return SumPooling()
        elif pooling == "max":
            return MaxPooling()
        else:
            raise ValueError(f"Pooling: {pooling} not supported !")

    def _embed_gin(self, dataset):
        """Embed molecules using GIN"""
        data_loader = DataLoader(
            dataset,
            batch_size=self.batch_size,
            collate_fn=dgl.batch,
            shuffle=False,
            drop_last=False,
        )

        mol_emb = []
        for batch_id, bg in enumerate(data_loader):
            if self.verbose:
                logger.debug("Processing batch {:d}/{:d}".format(batch_id + 1, len(data_loader)))
            nfeats = [
                bg.ndata.pop("atomic_number").to(torch.device("cpu")),
                bg.ndata.pop("chirality_type").to(torch.device("cpu")),
            ]
            efeats = [
                bg.edata.pop("bond_type").to(torch.device("cpu")),
                bg.edata.pop("bond_direction_type").to(torch.device("cpu")),
            ]
            with torch.no_grad():
                node_repr = self.featurizer(bg, nfeats, efeats)
            mol_emb.append(self._pooling_obj(bg, node_repr))
        mol_emb = torch.cat(mol_emb, dim=0).detach().cpu().numpy()
        return mol_emb

    def _embed_jtvae(self, dataset):
        """Embed molecules using JTVAE"""
        dataloader = DataLoader(dataset, batch_size=1, collate_fn=JTVAECollator(training=False))

        mol_emb = []
        for tree, tree_graph, mol_graph in dataloader:
            _, tree_vec, mol_vec = self.featurizer.encode(tree_graph, mol_graph)
            enc = torch.cat([tree_vec, mol_vec], dim=1).detach()
            mol_emb.append(enc)
        mol_emb = torch.cat(mol_emb, dim=0).cpu().numpy()
        return mol_emb

    def _embed(self, smiles: List[str], **kwargs):
        """Embed molecules into a latent space"""
        self._preload()
        dataset, successes = self.graph_featurizer(smiles, kind=self.kind)
        if self.kind in DGLModel.available_models(query="^jtvae"):
            mol_emb = self._embed_jtvae(dataset)
        else:
            mol_emb = self._embed_gin(dataset)

        mol_emb = list(mol_emb)
        out = []
        k = 0
        for success in successes:
            if success:
                out.append(mol_emb[k])
                k += 1
            else:
                out.append(None)
        return out

    @staticmethod
    def graph_featurizer(smiles: List[str], kind: Optional[str] = None):
        """
        Construct graphs from SMILES and featurize them

        Args:
            smiles: SMILES of molecules for embedding computation

        Returns:
            dataset: List of graphs constructed and featurized
            list of bool: Indicators for whether the SMILES string can be parsed by RDKit
        """
        if kind in DGLModel.available_models(query="^jtvae"):
            vocab = JTVAEVocab()

            tmp_file = tempfile.NamedTemporaryFile(delete=False)
            with fsspec.open(tmp_file.name, "w") as f:
                f.write("\n".join(smiles))
            dataset = JTVAEDataset(tmp_file.name, vocab, training=False)
            os.unlink(tmp_file.name)
            # JTVAE does not support failure
            success = [True] * len(smiles)
            if len(dataset) != len(smiles):
                raise ValueError("JTVAE failed to featurize some molecules !")
            return dataset, success

        else:
            graphs = []
            success = []
            for smi in smiles:
                try:
                    mol = dm.to_mol(smi)
                    if mol is None:
                        success.append(False)
                        continue
                    g = mol_to_bigraph(
                        mol,
                        add_self_loop=True,
                        node_featurizer=PretrainAtomFeaturizer(),
                        edge_featurizer=PretrainBondFeaturizer(),
                        canonical_atom_order=False,
                    )
                    graphs.append(g)
                    success.append(True)
                except Exception as e:
                    logger.error(e)
                    success.append(False)
            return graphs, success



class InvalidEntryPointTypeError(Exception):
    """Raised when an entry point is not of the correct type."""

    ...


class LoadingEntryPointError(Exception):
    """Raised when an entry point cannot be loaded."""

    ...


class MissingEntryPointError(Exception):
    """Raised when an entry point cannot be found."""

    ...


class MultipleEntryPointError(Exception):
    """Raised when multiple entry points are found for a given name."""

    ...



import pytest

import numpy as np
import datamol as dm

from molfeat._version import __version__ as MOLFEAT_VERSION
from molfeat.trans.fp import FPVecTransformer
from molfeat.trans.fp import FPVecFilteredTransformer
from molfeat.trans.base import MoleculeTransformer
from molfeat.trans.base import PrecomputedMolTransformer
from molfeat.trans.graph import AdjGraphTransformer
from molfeat.trans.graph import DGLGraphTransformer
from molfeat.trans.graph import TopoDistGraphTransformer
from molfeat.trans.graph import PYGGraphTransformer
from molfeat.trans.pretrained import PretrainedDGLTransformer
from molfeat.trans.pretrained import GraphormerTransformer
from molfeat.trans.pretrained import PretrainedHFTransformer

from molfeat.calc.atom import AtomCalculator
from molfeat.calc.bond import BondCalculator
from molfeat.calc.atom import AtomMaterialCalculator
from molfeat.calc import FPCalculator
from molfeat.calc import ScaffoldKeyCalculator
from molfeat.calc import RDKitDescriptors2D
from molfeat.calc import CATS
from molfeat.calc import Pharmacophore2D
from molfeat.calc._atom_bond_features import atom_chiral_tag_one_hot
from molfeat.calc._atom_bond_features import atom_one_hot
from molfeat.trans.graph import MolTreeDecompositionTransformer

from molfeat.utils.cache import MolToKey
from molfeat.utils.cache import FileCache
from molfeat.utils.state import compare_state


def _dummy_featurizer_fn(x):
    return np.random.random(1024)


FEATURIZERS_SPEC = {
    # fp
    "FPVecTransformer": lambda: FPVecTransformer(
        kind="ecfp:4",
        length=2044,
        useFeatures=True,
        dtype=np.float64,
    ),
    "MoleculeTransformerFP": lambda: MoleculeTransformer(
        FPCalculator("ecfp", radius=2),
        dtype=np.float64,
    ),
    "FPVecFilteredTransformer": lambda: FPVecFilteredTransformer(),
    # base
    "MoleculeTransformer": lambda: MoleculeTransformer(featurizer=_dummy_featurizer_fn),
    # pretrained
    "PretrainedDGLTransformer": lambda: PretrainedDGLTransformer(),
    "GraphormerTransformer": lambda: GraphormerTransformer(),
    "PretrainedHFTransformer": lambda: PretrainedHFTransformer(),
    # graph
    "AdjGraphTransformer_with_bonds": lambda: AdjGraphTransformer(
        atom_featurizer=AtomMaterialCalculator(),
        bond_featurizer=BondCalculator(),
    ),
    "AdjGraphTransformer_with_bonds_custom_atom": lambda: AdjGraphTransformer(
        atom_featurizer=AtomMaterialCalculator(
            featurizer_funcs={
                "atom_chiral_tag_one_hot": atom_chiral_tag_one_hot,
                "atom_one_hot": atom_one_hot,
            }
        ),
        bond_featurizer=BondCalculator(),
    ),
    "AdjGraphTransformer": lambda: AdjGraphTransformer(atom_featurizer=AtomCalculator()),
    "DGLGraphTransformer": lambda: DGLGraphTransformer(),
    "TopoDistGraphTransformer": lambda: TopoDistGraphTransformer(),
    "PYGGraphTransformer": lambda: PYGGraphTransformer(),
    "MolTreeDecompositionTransformer": lambda: MolTreeDecompositionTransformer(verbose=False),
}
FEATURIZERS_ATOM_PICKLES = ["AdjGraphTransformer_with_bonds_custom_atom"]
FEATURIZERS_BUILDER_ATOM_PICKLES = [FEATURIZERS_SPEC[k] for k in FEATURIZERS_ATOM_PICKLES]
FEATURIZERS_NAMES, FEATURIZERS_BUILDER = zip(
    *[(k, v) for k, v in FEATURIZERS_SPEC.items() if k not in FEATURIZERS_ATOM_PICKLES]
)


@pytest.mark.parametrize(
    "featurizer_builder",
    FEATURIZERS_BUILDER,
    ids=FEATURIZERS_NAMES,
)
def test_to_from_state(featurizer_builder):
    featurizer: MoleculeTransformer = featurizer_builder()

    # check to_state
    state = featurizer.to_state_dict()
    assert "name" in state
    assert "args" in state
    assert "_molfeat_version" in state

    # check from_state
    featurizer = MoleculeTransformer.from_state_dict(state)
    state2 = featurizer.to_state_dict()
    assert state == state2


@pytest.mark.parametrize(
    "featurizer_builder",
    FEATURIZERS_BUILDER,
    ids=FEATURIZERS_NAMES,
)
def test_to_from_state_yaml(featurizer_builder, tmp_path):
    featurizer: MoleculeTransformer = featurizer_builder()

    featurizer_path = tmp_path / "featurizer.yaml"

    featurizer.to_state_yaml_file(featurizer_path)
    state = featurizer.to_state_dict()

    featurizer = MoleculeTransformer.from_state_yaml_file(featurizer_path)
    state2 = featurizer.to_state_dict()
    assert state == state2


@pytest.mark.parametrize(
    "featurizer_builder",
    FEATURIZERS_BUILDER,
    ids=FEATURIZERS_NAMES,
)
def test_to_from_state_json(featurizer_builder, tmp_path):
    featurizer: MoleculeTransformer = featurizer_builder()

    featurizer_path = tmp_path / "featurizer.json"

    featurizer.to_state_json_file(featurizer_path)
    state = featurizer.to_state_dict()

    featurizer = MoleculeTransformer.from_state_json_file(featurizer_path)
    state2 = featurizer.to_state_dict()
    assert state == state2


@pytest.mark.parametrize(
    "featurizer_builder",
    FEATURIZERS_BUILDER_ATOM_PICKLES,
    ids=FEATURIZERS_ATOM_PICKLES,
)
def test_state_atom_bond_pickle(featurizer_builder, tmp_path):
    featurizer: MoleculeTransformer = featurizer_builder()

    # check to_state
    state = featurizer.to_state_dict()
    assert "name" in state
    assert "args" in state
    assert "_molfeat_version" in state

    # check from_state
    featurizer = MoleculeTransformer.from_state_dict(state)
    state2 = featurizer.to_state_dict()
    featurizer2 = MoleculeTransformer.from_state_dict(state2)
    to_remove = []
    for key in state["args"]:
        if key.endswith("is_pickled") and state["args"][key] == True:
            to_remove.append(key.replace("is_pickled", "").strip("_"))
    for key_val in to_remove:
        state["args"].pop(key_val, None)
        state2["args"].pop(key_val, None)
    assert state == state2
    mols = dm.freesolv().smiles.iloc[:10]
    out = featurizer(mols)
    out2 = featurizer2(mols)
    assert len(out) == len(out2) == len(mols)
    for o1, o2 in zip(out, out2):
        if isinstance(o1, (list, tuple)):
            assert np.all([np.allclose(i, j, atol=2) for i, j in zip(o1, o2)]) == True
        else:
            np.testing.assert_array_equal(o1, o2)


def test_PrecomputedMolTransformer_state(tmp_path):
    cache_path = tmp_path / "cache.parquet"
    smiles_list = ["[NH3+]CCSc1cc(-c2ccc[nH]2)c2c3c(ccc(F)c13)NC2=O", "CCCO"]

    featurizer = FPVecTransformer(kind="ecfp", length=2048)
    cache = FileCache(cache_file=cache_path, file_type="parquet", n_jobs=-1)

    featurizer_cache = PrecomputedMolTransformer(cache=cache, featurizer=featurizer)
    fps = featurizer_cache(smiles_list)

    # sanity check
    assert len(featurizer_cache.cache.cache) == 2

    # serialize to state
    state = featurizer_cache.to_state_dict()

    assert state["name"] == "PrecomputedMolTransformer"
    assert "base_featurizer" in state
    assert "cache" in state
    assert "_molfeat_version" in state

    # reload from state
    featurizer_cache2: PrecomputedMolTransformer = MoleculeTransformer.from_state_dict(state)  # type: ignore

    assert featurizer_cache2.base_featurizer.to_state_dict() == featurizer.to_state_dict()
    assert len(featurizer_cache2.cache.cache) == 2
    assert all(
        featurizer_cache2.cache.to_dataframe(pack_bits=True)
        == featurizer_cache2.cache.to_dataframe(pack_bits=True)
    )


def test_PrecomputedMolTransformer_init_from_state_file(tmp_path):
    cache_path = str(tmp_path / "cache.parquet")
    smiles_list = ["[NH3+]CCSc1cc(-c2ccc[nH]2)c2c3c(ccc(F)c13)NC2=O", "CCCO"]

    featurizer = FPVecTransformer(kind="ecfp", length=2048)
    cache = FileCache(cache_file=cache_path, file_type="parquet", n_jobs=-1)

    featurizer_cache = PrecomputedMolTransformer(cache=cache, featurizer=featurizer)
    fps = featurizer_cache(smiles_list)

    # sanity check
    assert len(featurizer_cache.cache.cache) == 2

    # serialize to state
    state = featurizer_cache.to_state_dict()

    # export yaml state file
    state_path = str(tmp_path / "state.yaml")

    featurizer_cache.to_state_yaml_file(filepath=state_path)

    assert state["name"] == "PrecomputedMolTransformer"
    assert "base_featurizer" in state
    assert "cache" in state
    assert "_molfeat_version" in state

    # reload from state
    featurizer_cache2: PrecomputedMolTransformer = PrecomputedMolTransformer(state_path=state_path)

    assert featurizer_cache2.base_featurizer.to_state_dict() == featurizer.to_state_dict()
    assert len(featurizer_cache2.cache.cache) == 2
    assert all(
        featurizer_cache2.cache.to_dataframe(pack_bits=True)
        == featurizer_cache2.cache.to_dataframe(pack_bits=True)
    )


def test_moltokey_state():
    hasher = MolToKey()
    state = hasher.to_state_dict()

    assert state == {"hash_name": "dm.unique_id"}

    hasher = MolToKey.from_state_dict(state)
    assert hasher.hash_name == "dm.unique_id"
    assert hasher.hash_fn == dm.unique_id


def test_fp_state():
    expected_res = [
        {
            "name": "FPCalculator",
            "module": "molfeat.calc.fingerprints",
            "args": {"length": 512, "method": "ecfp", "counting": False, "nBits": 512},
            "_molfeat_version": MOLFEAT_VERSION,
        },
        {
            "name": "FPCalculator",
            "module": "molfeat.calc.fingerprints",
            "args": {
                "length": 241,
                "method": "fcfp-count",
                "counting": True,
                "nBits": 241,
            },
            "_molfeat_version": MOLFEAT_VERSION,
        },
        {
            "name": "FPCalculator",
            "module": "molfeat.calc.fingerprints",
            "args": {"length": 2048, "method": "maccs", "counting": False},
            "_molfeat_version": MOLFEAT_VERSION,
        },
    ]
    for i, (fp_name, fp_len) in enumerate(zip(["ecfp", "fcfp-count", "maccs"], [512, 241, 2048])):
        calc = FPCalculator(fp_name, fp_len)
        state = calc.to_state_dict()
        assert state == expected_res[i]
    # check reloading
    calc = FPCalculator("ecfp", length=512, radius=3, useChirality=True)
    state = calc.to_state_dict()
    mol = dm.to_mol("C1CC(=O)NC(=O)[C@H]1N2C(=O)C3=CC=CC=C3C2=O")  # isomeric thalidomide
    expected = calc(mol)
    reloaded_calc = FPCalculator.from_state_dict(state)
    out = reloaded_calc(mol)
    np.testing.assert_array_almost_equal(expected, out)


def test_rdkit_descr_state():
    calc = RDKitDescriptors2D(replace_nan=True, augment=False)
    expected_state = {
        "name": "RDKitDescriptors2D",
        "module": "molfeat.calc.descriptors",
        "args": {
            "replace_nan": True,
            "augment": False,
            "descrs": None,
            "avg_ipc": True,
            "do_not_standardize": False,
        },
        "_molfeat_version": MOLFEAT_VERSION,
    }

    state = calc.to_state_dict()
    assert state == expected_state
    mol = dm.to_mol("Nc1cnn(-c2ccccc2)c(=O)c1Cl")
    expected = calc(mol)
    reloaded_calc = RDKitDescriptors2D.from_state_dict(state)
    out = reloaded_calc(mol)
    np.testing.assert_array_almost_equal(expected, out)


def test_pharmacophore_state():
    calc = Pharmacophore2D(factory="gobbi", length=1024, useCounts=False)
    expected_state = {
        "name": "Pharmacophore2D",
        "module": "molfeat.calc.pharmacophore",
        "args": {
            "factory": "gobbi",
            "useCounts": False,
            "minPointCount": 2,
            "maxPointCount": 3,
            "shortestPathsOnly": True,
            "includeBondOrder": False,
            "skipFeats": [],
            "trianglePruneBins": True,
            "bins": [(2, 3), (3, 4), (4, 5), (5, 6), (6, 7), (7, 8), (8, 100)],
            "length": 1024,
        },
        "_molfeat_version": MOLFEAT_VERSION,
    }

    state = calc.to_state_dict()
    assert state == expected_state
    mol = dm.to_mol("Nc1cnn(-c2ccccc2)c(=O)c1Cl")
    expected = calc(mol)
    reloaded_calc = Pharmacophore2D.from_state_dict(state)
    out = reloaded_calc(mol)
    np.testing.assert_array_almost_equal(expected, out)


def test_cats_state():
    calc = CATS(max_dist=8, scale="count", bins=[0, 3, 6, 8])
    expected_state = {
        "name": "CATS",
        "module": "molfeat.calc.cats",
        "args": {
            "max_dist": 8,
            "bins": [0, 3, 6, 8],
            "scale": "count",
            "use_3d_distances": False,
        },
        "_molfeat_version": MOLFEAT_VERSION,
    }

    state = calc.to_state_dict()
    assert state == expected_state
    mol = dm.to_mol("Nc1cnn(-c2ccccc2)c(=O)c1Cl")
    expected = calc(mol)
    reloaded_calc = CATS.from_state_dict(state)
    out = reloaded_calc(mol)
    np.testing.assert_array_almost_equal(expected, out)


def test_skeys_state():
    calc = ScaffoldKeyCalculator(normalize=True, use_scaffold=False, verbose=False)
    expected_state = {
        "name": "ScaffoldKeyCalculator",
        "module": "molfeat.calc.skeys",
        "args": {
            "normalize": True,
            "verbose": False,
            "use_scaffold": False,
        },
        "_molfeat_version": MOLFEAT_VERSION,
    }

    state = calc.to_state_dict()
    assert state == expected_state
    mol = dm.to_mol("Nc1cnn(-c2ccccc2)c(=O)c1Cl")
    expected = calc(mol)
    reloaded_calc = ScaffoldKeyCalculator.from_state_dict(state)
    out = reloaded_calc(mol)
    np.testing.assert_array_almost_equal(expected, out)


def test_filecache_state(tmp_path):
    # setup
    cache_path = tmp_path / "cache.parquet"
    cache = FileCache(cache_file=cache_path, file_type="parquet")
    cache(["CCC", "CCCCO"], featurizer=lambda x: [[0, 1, 1, 0], [0, 1, 1, 1]])

    # save state
    state = cache.to_state_dict(save_to_file=True)

    assert set(state.keys()) == {
        "_cache_name",
        "cache_file",
        "file_type",
        "mol_hasher",
        "n_jobs",
        "name",
        "clear_on_exit",
        "parquet_kwargs",
        "verbose",
    }

    assert dm.fs.exists(cache_path)

    # reload from state
    cache2 = FileCache.from_state_dict(state)
    assert len(cache2.cache) == 2
    assert all(cache.to_dataframe(pack_bits=True) == cache2.to_dataframe(pack_bits=True))


def test_state_comparison():
    state_dict1 = dict(_molfeat_version="0.1.2", value=2)
    state_dict2 = dict(_molfeat_version="0.1.2dev", value=2)
    state_dict3 = dict(_molfeat_version="0.1", value=2)
    state_dict4 = dict(value=2)
    state_dict5 = dict(_molfeat_version="1.2.3", value=100)
    state_dict6 = dict(_molfeat_version="0.1.2pre", value=0)

    expected_results = {
        "major": [True, True, True, False, False, False],
        1: [True, True, True, False, False, False],
        "micro": [True, True, False, False, False, False],
        None: [True, False, False, False, False, False],
    }
    for allowed_version in ["major", 1, "micro", None]:
        results = []
        results.append(compare_state(state_dict1, state_dict1, allow_version_level=allowed_version))
        results.append(compare_state(state_dict1, state_dict2, allow_version_level=allowed_version))
        results.append(compare_state(state_dict1, state_dict3, allow_version_level=allowed_version))
        results.append(compare_state(state_dict1, state_dict4, allow_version_level=allowed_version))
        results.append(compare_state(state_dict1, state_dict5, allow_version_level=allowed_version))
        results.append(compare_state(state_dict1, state_dict6, allow_version_level=allowed_version))
        assert expected_results[allowed_version] == results



from typing import List, Union
from typing import Callable
from typing import Optional

import numpy as np
import torch

from molfeat.utils import datatype
from molfeat.utils.pooler import Pooling
from molfeat.trans.base import MoleculeTransformer
from molfeat.utils.commons import _parse_to_evaluable_str
from molfeat.utils import requires

if requires.check("bio_embeddings"):
    from bio_embeddings import embed as bio_embedder


class ProtBioFingerprint(MoleculeTransformer):
    """
    Wrapper for general purpose biological sequence representations, as provided by [`bio_embeddings`](https://github.com/sacdallago/bio_embeddings)

    For a list of available embeddings, see: https://docs.bioembeddings.com/v0.2.2/api/bio_embeddings.embed.html

    !!! note:
        The embeddings proposed here are the general purpose embeddings, meaning that task-specific
        embeddings offered by `bio_embeddings` (e.g PBTucker, DeepBlast) are not included.

        According to the bio_embeddings documentation, `prottrans_bert_bfd` and `seqvec` are the best embeddings.
    """

    SUPPORTED_EMBEDDINGS = [
        "bepler",
        "cpcprot",
        "esm",
        "esm1b",
        "esm1v",
        "fasttext",
        "glove",
        "one_hot_encoding",
        "plus_rnn",
        "prottrans_albert_bfd",
        "prottrans_bert_bfd",
        "prottrans_t5_bfd",
        "prottrans_t5_uniref50",
        "prottrans_t5_xl_u50",
        "prottrans_xlnet_uniref100",
        "seqvec",
        "unirep",
        "word2vec",
    ]

    def __init__(
        self,
        featurizer: Union[str, Callable] = "seqvec",
        pooling: str = "mean",
        dtype: Callable = np.float32,
        device: Optional[Union[torch.device, str]] = None,
        layer_pooling: str = "sum",
        **kwargs,
    ):
        """Constructor for Deep Learning based Protein representation.
        SeqVec featurizer will e

        Args:
            featurizer: Name of callable of the embedding model
            pooling: Pooling method to use for sequence embedding. Defaults to "mean".
                If you set pooling to None, token representation will be returned
            dtype: Representation output datatype. Defaults to None.
            device: Torch device to move the model to. Defaults to None.
            layer_pooling: Layer-wise pooling method to use when > 1 layer exists. Default to 'sum'.
                If None, last layers is taken. This is relevant for `seqvec` mostly
        """
        if not requires.check("bio_embeddings"):
            raise ValueError(
                "Cannot use this featurizer without bio_embeddings (pip install 'bio_embeddings[all]')."
            )

        if isinstance(featurizer, bio_embedder.EmbedderInterface):
            featurizer = featurizer
            self._model_name = self.featurizer.name
        else:
            if (
                not isinstance(featurizer, str)
                or featurizer.lower() not in self.SUPPORTED_EMBEDDINGS
            ):
                raise ValueError("Unknown featurizer: {}".format(featurizer))
            self._model_name = featurizer.lower()
            featurizer = bio_embedder.name_to_embedder[self._model_name](device=device, **kwargs)

        super().__init__(featurizer=featurizer, dtype=dtype, **kwargs)
        self._fitted = True
        self._representation = "seq"
        self.pooling = Pooling(dim=0, name=pooling)
        self.layer_pooling = Pooling(dim=0, name=layer_pooling)
        if pooling is None:
            self._representation = "token"
        self._feat_length = None

    def __len__(self):
        """Get featurizer length"""
        return self.featurizer.embedding_dimension

    @property
    def n_layers(self):
        """Get the number of layers used in this embedding"""
        return self.featurizer.number_of_layers

    def __repr__(self):
        return "{}(model={}, pooling={}, dtype={})".format(
            self.__class__.__name__,
            _parse_to_evaluable_str(self._model_name),
            _parse_to_evaluable_str(self.pooling.name),
            _parse_to_evaluable_str(self.dtype),
        )

    def _pool(self, embedding: list):
        """Perform embedding pooling
        Args:
            embedding: input embedding
        """
        if self.n_layers > 1 and self.layer_pooling.name is not None:
            embedding = self.layer_pooling(embedding)
        if len(embedding.shape) > 2:
            # we forcefully take the last layers
            embedding = embedding[-1]
        return self.pooling(embedding)

    def _transform(
        self,
        protein_seq: str,
        **kwargs,
    ):
        """
        Transform a protein/nucleotide sequence into a feature vector.

        Args:
            protein: protein sequence as amino acid sequences

        Returns:
            Embedding of size (FEAT_DIM, N_LAYERS) for token embeddings
                and (FEAT_DIM, N_LAYERS) for sequence embeddings
        """

        rep = self.featurizer.embed(protein_seq)
        return self._pool(rep)

    def transform(self, seqs: List[str], names: Optional[List[str]] = None, **kwargs):
        """
        Transform a list of protein/nucleotide sequence into a feature vector.

        Args:
            seqs: list of protein/nucleotide sequence as amino acids
            names: names of the macromolecules.  Will be ignored
            kwargs: additional arguments for the featurizer

        Returns:
            Embedding of size (N_SEQS, FEAT_DIM) for token embeddings
                and (FEAT_DIM, N_LAYERS) for sequence embeddings
        """
        if not isinstance(seqs, list):
            seqs = [seqs]
        if isinstance(seqs[0], (list, tuple)) and len(seqs[0]) == 2:
            _, seqs = zip(*seqs)
            seqs = list(seqs)
        res = list(self.featurizer.embed_many(seqs, **kwargs))
        res = [self._pool(x) for x in res]
        return res

    def __call__(
        self,
        seqs: List[str],
        ignore_errors: bool = False,
        enforce_dtype: bool = True,
        **kwargs,
    ):
        r"""
        Compute molecular representation of a protein sequence.
        If ignore_error is True, a list of features and valid ids are returned.

        Args:
            seqs: list of protein or nucleotide sequence as amino acids
            enforce_dtype: whether to enforce the instance dtype in the generated fingerprint
            ignore_errors: Whether to ignore errors during featurization or raise an error.
            kwargs: Named parameters for the transform method

        Returns:
            feats: list of valid embeddings
            ids: all valid positions that did not failed during featurization.
                Only returned when ignore_errors is True.

        """
        features = self.transform(seqs, **kwargs)
        ids = np.arange(len(features))
        if ignore_errors:
            features, ids = self._filter_none(features)
        if self.dtype is not None and enforce_dtype:
            if self._representation.startswith("token"):
                features = [
                    datatype.cast(feat, dtype=self.dtype, columns=self.columns) for feat in features
                ]
            else:
                features = datatype.cast(features, dtype=self.dtype, columns=self.columns)
        if not ignore_errors:
            return features
        return features, ids



from typing import Callable
from typing import Union
from typing import Optional
from typing import Iterable

import torch
import numpy as np
import datamol as dm
import pandas as pd

from rdkit.DataStructs.cDataStructs import ConvertToExplicit
from rdkit.DataStructs.cDataStructs import SparseBitVect
from rdkit.DataStructs.cDataStructs import UIntSparseIntVect
from rdkit.DataStructs.cDataStructs import ExplicitBitVect
from rdkit.DataStructs.cDataStructs import CreateFromBitString


def ensure_explicit(x: Union[SparseBitVect, ExplicitBitVect]):
    """Ensure that the input vector is not a sparse bit vector

    Args:
        x: input vector

    Returns:
        converted: ExplicitBitVect if input is SparseBitVec, else input as is
    """
    if isinstance(x, SparseBitVect):
        x = ConvertToExplicit(x)
    return x


def to_sparse(x, dtype=None):
    r"""
    Converts dense tensor x to sparse format

    Args:
        x (torch.Tensor): tensor to convert
        dtype (torch.dtype, optional): Enforces new data type for the output.
            If None, it keeps the same datatype as x (Default: None)
    Returns:
        new torch.sparse Tensor
    """

    if dtype is not None:
        x = x.type(dtype)

    x_typename = torch.typename(x).split(".")[-1]
    sparse_tensortype = getattr(torch.sparse, x_typename)

    indices = torch.nonzero(x)
    if len(indices.shape) == 0:  # if all elements are zeros
        return sparse_tensortype(*x.shape)
    indices = indices.t()
    values = x[tuple(indices[i] for i in range(indices.shape[0]))]
    return sparse_tensortype(indices, values, x.size())


def to_tensor(x, gpu=False, dtype=None):
    r"""
    Convert a numpy array to tensor. The tensor type will be
    the same as the original array, unless specify otherwise

    Args:
        x (numpy.ndarray): Numpy array to convert to tensor type
        gpu (bool optional): Whether to move tensor to gpu. Default False
        dtype (torch.dtype, optional): Enforces new data type for the output

    Returns:
        New torch.Tensor
    """
    if not torch.is_tensor(x):
        try:
            if torch.is_tensor(x[0]):
                x = torch.stack(x)
        except:
            pass
        x = torch.as_tensor(x)
    if dtype is not None:
        x = x.to(dtype=dtype)
    if gpu and torch.cuda.is_available():
        x = x.cuda()
    return x


def to_numpy(x, copy=False, dtype=None):
    r"""
    Convert a tensor to numpy array.

    Args:
        x (Object): The Python object to convert.
        copy (bool, optional): Whether to copy the memory.
            By default, if a tensor is already on CPU, the
            Numpy array will be a view of the tensor.
        dtype (callable, optional): Optional type to cast the values to

    Returns:
        A new Python object with the same structure as `x` but where the tensors are now Numpy
        arrays. Not supported type are left as reference in the new object.
    """
    if isinstance(x, (list, tuple, np.ndarray)) and torch.is_tensor(x[0]):
        x = [to_numpy(xi, copy=copy, dtype=dtype) for xi in x]
    if isinstance(x, np.ndarray):
        pass
    elif torch.is_tensor(x):
        x = x.cpu().detach().numpy()
        x = x.copy()
    elif isinstance(x, SparseBitVect):
        tmp = np.zeros(x.GetNumBits(), dtype=int)
        for n_bit in list(x.GetOnBits()):
            tmp[n_bit] = 1
        x = tmp
    elif isinstance(x, ExplicitBitVect):
        x = dm.fp_to_array(x)
    elif hasattr(x, "GetNonzeroElements"):
        # one of the other rdkit type
        tmp = np.zeros(x.GetLength())
        bit_idx, values = np.array(list(x.GetNonzeroElements().items())).T
        tmp[bit_idx] = values
        x = tmp
    else:
        x = np.asarray(x)
    if dtype is not None:
        x = x.astype(dtype)
    return x


def to_fp(arr: np.ndarray, bitvect: bool = True, sparse: bool = False):
    """Convert numpy array to fingerprint

    Args:
        arr: Numpy array to convert to bitvec
        bitvect: whether to assume the data is a bitvect or intvect
        sparse: whether to convert to sparse bit vect

    Returns:
        fp: RDKit bit vector
    """
    if not isinstance(arr, list) and arr.ndim > 1:
        raise ValueError("Expect a 1D array as input !")
    if not bitvect:
        fp = UIntSparseIntVect(len(arr))
        for ix, value in enumerate(arr):
            fp[ix] = int(value)
    elif sparse:
        onbits = np.where(arr == 1)[0].tolist()
        fp = SparseBitVect(arr.shape[0])
        fp.SetBitsFromList(onbits)
    else:
        arr = np.asarray(arr)
        bitstring = "".join(arr.astype(str))
        fp = CreateFromBitString(bitstring)
    return fp


def is_dtype_tensor(dtype):
    r"""
    Verify if the dtype is a torch dtype

    Args:
        dtype (callable): The dtype of a value. E.g. np.int32, str, torch.float

    Returns:
        A boolean saying if the dtype is a torch dtype
    """
    return isinstance(dtype, torch.dtype) or (dtype == torch.Tensor)


def is_dtype_bitvect(dtype):
    """
    Verify if the dtype is a bitvect type

    Args:
        dtype (callable): The dtype of a value. E.g. np.int32, str, torch.float

    Returns:
        A boolean saying if the dtype is a torch dtype
    """
    return dtype in [ExplicitBitVect, SparseBitVect] or isinstance(
        dtype, (ExplicitBitVect, SparseBitVect)
    )


def is_dtype_numpy(dtype):
    r"""
    Verify if the dtype is a numpy dtype

    Args:
        dtype (callable): The dtype of a value. E.g. np.int32, str, torch.float
    Returns
        A boolean saying if the dtype is a numpy dtype
    """
    is_torch = is_dtype_tensor(dtype)
    is_num = dtype in (int, float, complex)
    if hasattr(dtype, "__module__"):
        is_numpy = dtype.__module__ == "numpy"
    else:
        is_numpy = False
    return (is_num or is_numpy) and not is_torch


def is_null(obj):
    """Check if an obj is null (nan, None or array of nan)"""
    array_nan = False
    all_none = False
    try:
        tmp = to_numpy(obj)
        array_nan = np.all(np.isnan(tmp))
    except:
        pass
    try:
        all_none = all(x is None for x in obj)
    except:
        pass
    return obj is None or all_none or array_nan


def cast(fp, dtype: Optional[Callable] = None, columns: Optional[Iterable] = None):
    """Change the datatype of a list of input array

    Args:
        fp (array): Input array to cast (2D)
        dtype: datatype to cast to
        columns: column names for pandas dataframe
    """
    if fp is None or dtype is None:
        return fp
    if isinstance(fp, dict):
        fp = {k: cast(v, dtype=dtype, columns=columns) for k, v in fp.items()}
    elif dtype in [tuple, list]:
        fp = list(fp)
    elif is_dtype_numpy(dtype):
        if isinstance(fp, (list, tuple)) and not np.isscalar(fp[0]):
            fp = [to_numpy(fp_i, dtype=dtype) for fp_i in fp]
            fp = to_numpy(fp, dtype=dtype)
        else:
            fp = to_numpy(fp, dtype=dtype)
    elif is_dtype_tensor(dtype):
        if isinstance(fp, (list, tuple)) and not np.isscalar(fp[0]):
            tmp_fp = to_numpy(fp[0])
            if len(tmp_fp.shape) > 1:
                fp = torch.cat([to_tensor(fp_i, dtype=dtype) for fp_i in fp])
            else:
                fp = torch.stack([to_tensor(fp_i, dtype=dtype) for fp_i in fp])
        else:
            fp = to_tensor(fp, dtype=dtype)
    elif dtype in [pd.DataFrame, "dataframe", "pandas", "df"]:
        fp = [feat if feat is not None else [] for feat in fp]
        fp = pd.DataFrame(fp)
        if columns is not None:
            fp.columns = columns
    elif is_dtype_bitvect(dtype):
        fp = [to_fp(feat, sparse=(dtype == SparseBitVect)) for feat in fp]
    else:
        raise TypeError("The type {} is not supported".format(dtype))
    return fp



import unittest as ut
import pytest

import tempfile

import joblib

import datamol as dm
import numpy as np

from molfeat.calc import CATS
from molfeat.calc import ScaffoldKeyCalculator
from molfeat.calc import RDKitDescriptors2D
from molfeat.calc import RDKitDescriptors3D
from molfeat.calc import MordredDescriptors
from molfeat.calc import USRDescriptors
from molfeat.calc import ElectroShapeDescriptors
from molfeat.calc.skeys import skdistance
from molfeat.utils import requires


class TestDescPharm(ut.TestCase):
    r"""Test cases for descriptors and pharmacophore generation"""
    smiles = [
        "CCOc1c(OC)cc(CCN)cc1OC",
        "COc1cc(CCN)cc(OC)c1OC",
        "C[C@@H]([NH3+])Cc1c2ccoc2c(Br)c2ccoc12",
    ]
    EXTRA_LARGE_MOL = "CC(C)CC(NCCNC(=O)C(CCC(O)=O)NC(C)=O)C(=O)NC(Cc1ccc(O)cc1)C(=O)NC(CC(C)C)C(=O)NC(C(C)C)C(=O)NC(C)C(=O)NCC(=O)NC(CCC(O)=O)C(=O)NC(CCCNC(N)=N)C(=O)NCC(=O)NC(Cc1ccccc1)C(=O)NC(Cc1ccccc1)C(=O)NC(Cc1ccc(O)cc1)C(=O)NC(C(C)O)C(=O)N1CCCC1C(=O)NC(C)C(O)=O"

    def test_rdkit2d(self):
        calc = RDKitDescriptors2D()
        fps = calc(self.smiles[0])
        self.assertEqual(len(fps), len(calc))

        sm = "CC[C@@H]1[C@@H]2C[C@H](O)CC[C@]2(C)[C@H]2CC[C@@]3(C)[C@@H](CC[C@@H]3[C@H](C)CCOS(=O)(=O)O[Na])[C@@H]2[C@@H]1O"
        sm_disconnected = "CC[C@@H]1[C@@H]2C[C@H](O)CC[C@]2(C)[C@H]2CC[C@@]3(C)[C@@H](CC[C@@H]3[C@H](C)CCOS(=O)(=O)[O-])[C@@H]2[C@@H]1O.[Na+]"
        # force sanitization would return same descriptors for both of this
        fps = calc(sm)
        fps2 = calc(sm_disconnected)
        np.testing.assert_allclose(fps, fps2)

        # with the fix we should not have any value that is nan after sanitization
        # neither for Charge related or BCut2D properties
        self.assertFalse(np.isnan(fps).any())

        # we should have nan values at all bcut columns
        # if we do not standardize
        calc_nan = RDKitDescriptors2D(do_not_standardize=True)
        fps = calc_nan(sm)
        bcut_cols = [i for i, x in enumerate(calc.columns) if "bcut" in x.lower()]
        self.assertTrue(np.isnan(fps[bcut_cols]).all())

        # sanity check for large molecules
        ipc_colums = calc.columns.index("Ipc")
        fps = calc(self.EXTRA_LARGE_MOL)
        self.assertLessEqual(fps[ipc_colums], 1e3)

    def test_rdkit3d(self):
        calc = RDKitDescriptors3D()
        mol = dm.conformers.generate(dm.to_mol(self.smiles[0]))
        fps = calc(mol)
        self.assertEqual(len(fps), len(calc))

    @pytest.mark.xfail(not requires.check("mordred"), reason="3rd party module mordred is missing")
    def test_mordred(self):
        calc = MordredDescriptors()
        fps = calc(self.smiles[0])
        self.assertEqual(len(fps), len(calc))

    def test_cats_2d(self):
        smiles = "Nc1cnn(-c2ccccc2)c(=O)c1Cl"
        mol = dm.to_mol(smiles)

        calc = CATS(max_dist=5, scale="raw", use_3d_distances=False)
        assert calc(mol).shape == (126,)

    def test_cats_3d_missing_conformer(self):
        smiles = "Nc1cnn(-c2ccccc2)c(=O)c1Cl"
        mol = dm.to_mol(smiles)

        calc = CATS(max_dist=5, scale="raw", use_3d_distances=True)

        with pytest.raises(ValueError):
            calc(mol)

    def test_cats_3d(self):
        smiles = "Nc1cnn(-c2ccccc2)c(=O)c1Cl"
        mol = dm.to_mol(smiles)
        mol = dm.conformers.generate(mol)

        calc = CATS(max_dist=5, scale="raw", use_3d_distances=True)
        assert calc(mol).shape == (126,)

    def test_cats_pickle(self):
        smiles = "Nc1cnn(-c2ccccc2)c(=O)c1Cl"
        mol = dm.to_mol(smiles)

        calc = CATS(max_dist=5, scale="raw", use_3d_distances=False)

        # compute fp
        fp1 = calc(mol)

        fpath = tempfile.NamedTemporaryFile().name

        # pickle
        joblib.dump(calc, fpath)

        # unpickle
        calc = joblib.load(fpath)

        # compute fp
        fp2 = calc(mol)

        # check
        assert np.allclose(fp1, fp2)

    def test_shape_descriptors(self):
        calc = USRDescriptors("usrcat")
        with self.assertRaises(ValueError) as context:
            calc(self.smiles[0])
        mol_with_conf = dm.conformers.generate(dm.to_mol(self.smiles[0]))
        out = calc(mol_with_conf)
        self.assertEqual(out.shape[-1], len(calc))

        calc2 = ElectroShapeDescriptors("mmff94")
        out2 = calc2(mol_with_conf)
        self.assertEqual(out2.shape[-1], len(calc2))

    def test_scaffkey(self):
        calc = ScaffoldKeyCalculator()
        fps = calc(self.smiles[0])
        columns = calc.columns
        col_to_check = [
            "n_atom_in_rings",
            "n_nitrogen",
            "n_heteroatoms",
            "n_ring_system",
            "n_carbon_het_carbon_het_bonds",
        ]
        expected_output = [6, 1, 4, 1, 2]
        comp_res = [fps[columns.index(x)] for x in col_to_check]
        self.assertEqual(expected_output, comp_res)

    def test_scaff_skdist(self):
        calc = ScaffoldKeyCalculator()
        smiles = dm.freesolv()["smiles"][:10].tolist()
        fps = [calc(x) for x in smiles]
        fps1 = np.asarray(fps[:6])
        fps2 = np.asarray(fps[6:])
        # compute batch
        pairwise_dist = skdistance(fps1, fps2, cdist=True)
        # compute singletons
        dist = []
        for i in range(fps2.shape[0]):
            dist.append(skdistance(fps[0], fps2[i, :], cdist=False))

        np.testing.assert_allclose(np.asarray(dist), pairwise_dist[0, :])


if __name__ == "__main__":
    ut.main()



from typing import Callable
from typing import List
from typing import Optional
from typing import Union

import re
import copy
import numpy as np
from rdkit.Chem import rdchem

from molfeat.calc import get_calculator, FP_FUNCS
from molfeat.trans.base import MoleculeTransformer
from molfeat.utils import datatype
from molfeat.utils.commons import _parse_to_evaluable_str


_UNSERIALIZABLE_FPS = []


class FPVecTransformer(MoleculeTransformer):
    r"""
    Molecular fingerprinter that computes various fingerprints and descriptors regularly used in QSAR modeling.

    !!! note
        For fingerprints with a radius parameter, you can provide the radius using the notation "fp:radius".
        For example "Morgan Circular 2" can be written as "morgan:2". Under the hood, morgan and ecfp fingerprints
        are equated with the proper radius/diameter adjustment.

        For counting fingerprints, you just need to add the '-count' suffix to the name of the fingerprint. For example:
        "morgan-count:2"
    """

    AVAILABLE_FPS = list(FP_FUNCS.keys()) + [
        "desc3D",
        "desc2D",
        "mordred",
        "cats2D",
        "cats3D",
        "pharm2D",
        "pharm3D",
        "scaffoldkeys",
        "skeys",
    ]

    def __init__(
        self,
        kind: str = "ecfp:4",
        length: int = 2000,
        n_jobs: int = 1,
        verbose: bool = False,
        dtype: Callable = np.float32,
        parallel_kwargs: Optional[dict] = None,
        **params,
    ):
        """Molecular to vector fingerprinter

        Args:
            kind (str, optional): Name of the fingerprints (one supported fingerprints: see self.AVAILABLE_FPS). Defaults to "ecfp4".
            length (int, optional): Length of the fingerprint. Defaults to 2000.
            n_jobs (int, optional): Number of jobs. Defaults to 1.
            verbose (bool, optional): Verbosity level. Defaults to False.
            dtype (Callable, optional): Data type. Defaults to np.float32.
            parallel_kwargs (dict, optional): Optional arguments to pass to dm.parallelized when required. Defaults to None.
            params (dict, optional): Any additional parameters to the fingerprint function
        """
        self._save_input_args()

        # remove any featurizer that was passed as argument
        params.pop("featurizer", None)
        self._feat_params = params
        featurizer = self._prepare_featurizer(kind, length, **params)
        super().__init__(
            featurizer=featurizer,
            n_jobs=n_jobs,
            verbose=verbose,
            dtype=dtype,
            parallel_kwargs=parallel_kwargs,
            **params,
        )
        self.kind = kind
        self.length = length
        self._length = None
        # update length for featurizer that have they fixed length
        # EN: setting up a protected _length function helps to bypass
        # the absurd "is" comparison done by sklearn in clone
        # note that the featurizer length would likely be ignored by featurizer
        # that do not support a variable length
        if hasattr(self.featurizer, "__len__"):
            self._length = len(featurizer)
        self._input_params.update(kind=kind, length=length)
        if self.kind.lower() in _UNSERIALIZABLE_FPS:
            self.parallel_kwargs.update(scheduler="threads")

    def __len__(self):
        """Compute featurizer length"""
        if getattr(self, "cols_to_keep", None) is None and self._length is not None:
            return self._length
        return super().__len__()

    def _get_param_names(self):
        """Get parameter names for the estimator"""
        out = self._input_params.keys()
        return [x for x in out if x != "featurizer"]

    @classmethod
    def _prepare_featurizer(cls, kind: str, length: int, **params):
        """Prepare featurizer from its name and parameters

        Args:
            kind: Name of the featurizer
            length: Length of the featurizer
        Returns:
            calculator (Callable): fingerprint calculator
        """
        match = re.search(r":(\d+)$", kind)
        radius = None
        if match is not None:
            radius = match.group(1)
        if radius is not None:
            kind = kind.replace(radius, "").strip(":").lower()
            radius = int(radius)
            if any(x in kind for x in ["ecfp", "fcfp"]):
                radius = max(radius // 2, 1)
            params["radius"] = radius
        if any(x in kind for x in ["morgan", "morgan_circular", "morgan-circular"]):
            kind.replace("_circular", "").replace("-circular", "").replace("morgan", "ecfp")
        if kind not in cls.AVAILABLE_FPS:
            raise ValueError(f"{kind} is not a valid featurizer")
        params["length"] = length

        return get_calculator(kind, **params)

    def _update_params(self):
        params = copy.deepcopy(self._input_params)
        params.pop("featurizer", None)
        params.pop("length", None)
        params.pop("kind", None)
        params.pop("verbose", None)
        params.pop("dtype", None)
        params.pop("n_jobs", None)
        self._fitted = False
        self.featurizer = self._prepare_featurizer(self.kind, self.length, **params)

    def __repr__(self):
        return "{}(kind={}, length={}, dtype={})".format(
            self.__class__.__name__,
            _parse_to_evaluable_str(self.kind),
            _parse_to_evaluable_str(self.length),
            _parse_to_evaluable_str(self.dtype),
        )

    def __str__(self):
        # The output for the print function
        return self.__repr__()

    def __eq__(self, other):
        same_type = type(self) == type(other)
        return same_type and all(
            [getattr(other, k) == v for k, v in self.get_params() if not callable(v)]
        )

    def __ne__(self, other):
        return not (self == other)

    def __hash__(self):
        return hash(repr(self))


class FPVecFilteredTransformer(FPVecTransformer):
    r"""
    Fingerprint molecule transformer with columns filters applying to the featurized vector when `fit` is called
    """

    def __init__(
        self,
        kind: str = "ecfp:4",
        length: int = 2000,
        occ_threshold: float = 0,
        del_invariant: bool = False,
        n_jobs: int = 1,
        verbose: bool = False,
        dtype: Callable = np.float32,
        **params,
    ):
        """Molecular to vector featurization with filtering applied

        Args:
            kind (str, optional): Name of the fingerprints (one supported fingerprints: see self.AVAILABLE_FPS). Defaults to "ecfp4".
            length (int, optional): Length of the fingerprint. Defaults to 2000.
            occ_threshold (float, optional): Minimum proportion a columns need to be non null to be kept.
            del_invariant (bool, optional): Whether to delete columns that are invariant.
            n_jobs (int, optional): Number of jobs. Defaults to 1.
            verbose (bool, optional): Verbosity level. Defaults to False.
            dtype (Callable, optional): Data type. Defaults to np.float32.
            params (dict, optional): Any additional parameters to the fingerprint function
        """

        super().__init__(
            kind=kind,
            length=length,
            n_jobs=n_jobs,
            verbose=verbose,
            dtype=dtype,
            **params,
        )
        self.occ_threshold = occ_threshold
        self.del_invariant = del_invariant
        self._input_params.update(occ_threshold=occ_threshold, del_invariant=del_invariant)

    def _update_params(self):
        params = copy.deepcopy(self._input_params)
        params.pop("featurizer", None)
        params.pop("length", None)
        params.pop("kind", None)
        params.pop("verbose", None)
        params.pop("dtype", None)
        params.pop("n_jobs", None)
        params.pop("occ_threshold", None)
        params.pop("del_invariant", None)
        self.featurizer = self._prepare_featurizer(self.kind, self.length, **params)

    def __repr__(self):
        return "{} (kind={}, length={}, occ_threshold={}, del_invariant={}, dtype={})".format(
            self.__class__.__name__,
            _parse_to_evaluable_str(self.kind),
            _parse_to_evaluable_str(self.length),
            _parse_to_evaluable_str(self.occ_threshold),
            _parse_to_evaluable_str(self.del_invariant),
            _parse_to_evaluable_str(self.dtype),
        )

    def fit(self, X: List[Union[rdchem.Mol, str]], y: Optional[list] = None, **fit_params):
        """Fit the current transformer on given dataset.

        The goal of fitting is for example to identify nan columns values
        that needs to be removed from the dataset

        Args:
            X: input list of molecules
            y (list, optional): Optional list of molecular properties. Defaults to None.

        Returns:
            self: MolTransformer instance after fitting
        """

        feats = self.transform(X, ignore_errors=True)
        lengths = [len(x) for x in feats if not datatype.is_null(x)]

        if lengths:
            # we will ignore all nan
            feats = datatype.to_numpy([f for f in feats if not datatype.is_null(f)])
            # all nan columns
            unwanted_columns = []
            # let's adjsut occ to float
            occ_threshold = self.occ_threshold
            if occ_threshold > 1:
                occ_threshold = occ_threshold / feats.shape[0]
            # not nan
            unwanted_columns.append(~np.any(np.isnan(feats), axis=0))
            # not enough set bits
            unwanted_columns.append(
                (np.count_nonzero(feats, axis=0) / feats.shape[0]) > occ_threshold
            )
            if self.del_invariant:
                unwanted_columns.append(~np.all(feats == feats[0, :], axis=0))
            self.cols_to_keep = (np.logical_and.reduce(unwanted_columns)).nonzero()[0]
        self._fitted = True
        return self



import importlib.resources as pkg_resources
import pandas as pd


def get_df(name: str):
    """Get a static dataframe file in this folder

    Args:
        name: name of the data to get

    Returns:
        data (pd.DataFrame): Loaded data
    """
    if not name.endswith(".xz"):
        name = f"{name}.xz"
    with pkg_resources.open_binary("molfeat.data", name) as instream:
        data = pd.read_pickle(instream, compression="xz")
    return data



from rdkit import Chem

PERIODIC_TABLE = Chem.GetPeriodicTable()

ATOM_LIST = [
    "C",
    "N",
    "O",
    "S",
    "F",
    "Si",
    "P",
    "Cl",
    "Br",
    "Mg",
    "Na",
    "Ca",
    "Fe",
    "As",
    "Al",
    "I",
    "B",
    "V",
    "K",
    "Tl",
    "Yb",
    "Sb",
    "Sn",
    "Ag",
    "Pd",
    "Co",
    "Se",
    "Ti",
    "Zn",
    "H",
    "Li",
    "Ge",
    "Cu",
    "Au",
    "Ni",
    "Cd",
    "In",
    "Mn",
    "Zr",
    "Cr",
    "Pt",
    "Hg",
    "Pb",
]
ATOM_NUM_H = [0, 1, 2, 3, 4]
IMPLICIT_VALENCE = [0, 1, 2, 3, 4, 5, 6]
CHARGE_LIST = [-3, -2, -1, 0, 1, 2, 3]
RADICAL_E_LIST = [0, 1, 2]
HYBRIDIZATION_LIST = [
    Chem.rdchem.HybridizationType.names[k]
    for k in sorted(Chem.rdchem.HybridizationType.names.keys(), reverse=True)
    if k != "OTHER"
]
ATOM_DEGREE_LIST = range(5)
CHIRALITY_LIST = ["R"]  # alternative is just S
BOND_TYPES = [
    Chem.rdchem.BondType.SINGLE,
    Chem.rdchem.BondType.DOUBLE,
    Chem.rdchem.BondType.TRIPLE,
    Chem.rdchem.BondType.AROMATIC,
]
BOND_STEREO = [0, 1, 2, 3, 4, 5]

HARD_ATOM_LIMIT = 119
ATOM_ALPHABET = [PERIODIC_TABLE.GetElementSymbol(i) for i in range(1, HARD_ATOM_LIMIT)]
AMINO_ACID_ALPHABET = list("ARNDCQEGHILKMFPSTWYVBZX*")
BOND_ALPHABET = list(
    Chem.BondType().names.values()
)  # if you want the names: use temp_bond_name.names.keys()
SMILES_ALPHABET = list("#%)(+*-/.1032547698:=@[]\\consb") + ATOM_ALPHABET + ["se"]



from typing import Union

import datamol as dm
from rdkit.Chem.rdMHFPFingerprint import MHFPEncoder
from rdkit.Chem import rdchem


def MHFP(
    x: Union[rdchem.Mol, str],
    n_permutations: int = 128,
    radius: int = 3,
    min_radius: int = 1,
    rings: bool = True,
    isomeric: bool = False,
    kekulize: bool = False,
    seed: int = 0,
    **kwargs,
):
    """Compute Unfolded MHFP fingerprint from rdkit

    Args:
        x: input molecule
        n_permutations (int, optional): Number of permutations of the fp. Defaults to 512.
        radius (int, optional): Radius of the fingerprint. Defaults to 3.
        min_radius (int, optional): Minimum radius to consider. Defaults to 1.
        rings (bool, optional): Whether to consider rings
        isomeric (bool, optional): Whether to use isomeric variant. Defaults to False.
        kekulize (bool, optional): Whether to kekulize molecule first. Defaults to False.
        seed (int, optional): optional seed to use. Default to 0

    Returns:
        fp: fingerprint
    """
    if isinstance(x, str):
        x = dm.to_mol(x)
    mhfp_encoder = MHFPEncoder(n_permutations, seed)
    encoded_fp = mhfp_encoder.EncodeMol(
        x,
        radius=radius,
        rings=rings,
        isomeric=isomeric,
        kekulize=kekulize,
        min_radius=min_radius,
    )
    return encoded_fp


def SECFP(
    x: Union[rdchem.Mol, str],
    n_permutations: int = 128,
    nBits: int = 2048,
    radius: int = 3,
    min_radius: int = 1,
    rings: bool = True,
    isomeric: bool = False,
    kekulize: bool = False,
    seed: int = 0,
    **kwargs,
):
    """Compute SECFP (folded MHFP) fingerprint

    Args:
        x: input molecule
        n_permutations (int, optional): Number of permutations of the fp. Defaults to 512.
        nBits (int, optional): Length of the fingerprint. Defaults to 2048.
            We use nBits to be consistent with other fingerprints.
        radius (int, optional): Radius of the fingerprint. Defaults to 3.
        min_radius (int, optional): Minimum radius to consider. Defaults to 1.
        rings (bool, optional): Whether to consider rings
        isomeric (bool, optional): Whether to use isomeric variant. Defaults to False.
        kekulize (bool, optional): Whether to kekulize molecule first. Defaults to False.
        seed (int, optional): optional seed to use. Default to 0

    Returns:
        fp: fingerprint
    """
    if isinstance(x, str):
        x = dm.to_mol(x)
    mhfp_encoder = MHFPEncoder(n_permutations, seed)
    encoded_fp = mhfp_encoder.EncodeSECFPMol(
        x,
        radius=radius,
        rings=rings,
        isomeric=isomeric,
        kekulize=kekulize,
        min_radius=min_radius,
        length=nBits,
    )
    return encoded_fp



import unittest as ut
import pytest

import datamol as dm
import numpy as np

try:
    from dgllife.utils import CanonicalAtomFeaturizer
    from dgllife.utils import CanonicalBondFeaturizer
    from dgllife.utils import WeaveAtomFeaturizer
    from dgllife.utils import WeaveEdgeFeaturizer
except ImportError:
    pass

from molfeat.calc.atom import AtomCalculator
from molfeat.calc.atom import DGLCanonicalAtomCalculator
from molfeat.calc.atom import DGLWeaveAtomCalculator
from molfeat.calc.bond import BondCalculator
from molfeat.calc.bond import DGLCanonicalBondCalculator
from molfeat.calc.bond import EdgeMatCalculator
from molfeat.calc.bond import DGLWeaveEdgeCalculator
from molfeat.calc._atom_bond_features import bond_type_one_hot
from molfeat.calc._atom_bond_features import bond_is_in_ring
from molfeat.calc._atom_bond_features import atom_degree_one_hot
from molfeat.calc._atom_bond_features import atom_one_hot
from molfeat.calc._atom_bond_features import atom_implicit_valence_one_hot

from molfeat.utils import requires


CUSTOM_BOND_FEATURIZER = {
    "bond_type_one_hot": bond_type_one_hot,
    "bond_is_in_ring": bond_is_in_ring,
    "bond_custom_val": lambda x: [0, 0],
}
CUSTOM_ATOM_FEATURIZER = {
    "atom_one_hot": atom_one_hot,
    "atom_degree_one_hot": atom_degree_one_hot,
    "atom_custom_val_1": lambda x: [0, 0],
    "atom_custom_val_2": atom_implicit_valence_one_hot,
}
CALCULATOR_SPECS = {
    # fp
    "AtomCalculator": lambda: AtomCalculator(),
    "WeaveAtomCalculator": lambda: DGLWeaveAtomCalculator(),
    "BondCalculator": lambda: BondCalculator(),
    "EdgeMatCalculator": lambda: EdgeMatCalculator(),
    "CustomBondCalculator": lambda: BondCalculator(featurizer_funcs=CUSTOM_BOND_FEATURIZER),
    "CustomAtomCalculator": lambda: AtomCalculator(featurizer_funcs=CUSTOM_ATOM_FEATURIZER),
}


@pytest.mark.parametrize(
    "calculator_builder",
    CALCULATOR_SPECS.values(),
    ids=CALCULATOR_SPECS.keys(),
)
def test_to_from_state(calculator_builder):
    calculator = calculator_builder()
    # check to_state
    state = calculator.to_state_dict()
    assert "name" in state
    assert "args" in state
    assert "_molfeat_version" in state

    # check from_state
    calculator2 = calculator.__class__.from_state_dict(state)
    state2 = calculator2.to_state_dict()
    assert state == state2


@pytest.mark.xfail(not requires.check("dgllife"), reason="3rd party module dgllife is missing")
class TestGraphCalculator(ut.TestCase):
    r"""Test cases for basic graph featurizer vs dgl generation"""
    smiles = [
        "CCOc1c(OC)cc(CCN)cc1OC",
        "COc1cc(CCN)cc(OC)c1OC",
        "C[C@@H]([NH3+])Cc1c2ccoc2c(Br)c2ccoc12",
        "C.C1CCC1",
        "N",
    ]
    mols = [dm.to_mol(x) for x in smiles]

    def test_atom_calculator(self):
        atom_calc = AtomCalculator()
        dgl_atom_calc = DGLCanonicalAtomCalculator()
        true_dgl_atom_calc = CanonicalAtomFeaturizer(atom_data_field="hv")
        for mol in self.mols:
            a1 = atom_calc(mol)
            a2 = dgl_atom_calc(mol)
            a3 = true_dgl_atom_calc(mol)
            np.testing.assert_allclose(a2["hv"], a3["hv"])

    def test_bond_calculator(self):
        bond_calc = BondCalculator()
        dgl_bond_calc = DGLCanonicalBondCalculator()
        true_dgl_bond_calc = CanonicalBondFeaturizer(bond_data_field="he")
        for mol in self.mols:
            b1 = bond_calc(mol)
            b2 = dgl_bond_calc(mol)
            b3 = true_dgl_bond_calc(mol)
            if "he" in b3:
                np.testing.assert_allclose(b2["he"], b3["he"])

    def test_weave_atom_calculator(self):
        true_dgl_weave_calc = WeaveAtomFeaturizer(atom_data_field="hv")
        dgl_weave_calc = DGLWeaveAtomCalculator()
        for mol in self.mols:
            b1 = true_dgl_weave_calc(mol)
            b2 = dgl_weave_calc(mol)
            np.testing.assert_allclose(b1["hv"], b2["hv"])

    def test_weave_bond_calculator(self):
        true_dgl_weave_calc = WeaveEdgeFeaturizer(edge_data_field="he")
        dgl_weave_calc = DGLWeaveEdgeCalculator()
        for mol in self.mols:
            b1 = true_dgl_weave_calc(mol)
            b2 = dgl_weave_calc(mol)
            np.testing.assert_allclose(b1["he"], b2["he"])


if __name__ == "__main__":
    ut.main()
